# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLAB (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"

import geospacelab.datahub.sources.madrigal.utilities as utilities

from geospacelab.datahub import DatabaseModel


class FMIDatabase(DatabaseModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj


fmi_database = FMIDatabase('FMI')
fmi_database.url = 'https://space.fmi.fi/'
fmi_database.category = 'online database'
fmi_database.Notes = ''



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import numpy as np

from geospacelab.datahub.sources.esa_eo.swarm.loader import LoaderModel

# define the default variable name dictionary
default_variable_name_dict = {
    'CDF_EPOCH': 't',
    'SC_GEO_LAT': 'Latitude',
    'SC_GEO_LON': 'Longitude',
    'SC_QD_LAT': 'Latitude_QD',
    'SC_QD_LON': 'Longitude_QD',
    'J': 'J',
    'J_E_QD': 'J_QD',
    't_Q': 't_qual',
    'RMS_MISFIT': 'RMS_misfit',
    'CONFIDENCE': 'Confidence',
}


class Loader(LoaderModel):
    """
    Load SWARM 2Hz or 16HZ TII data products. Currently support versions higher than "0301".

    The class is a hierarchy of :class:`SWARM data LoaderModel <geospacelab.datahub.sources.esa_eo.swarm.loader.LoaderModel>`

    """
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('variable_name_dict', default_variable_name_dict)
        super(Loader, self).__init__(*args, **kwargs)

    def load_data(self, **kwargs):
        super(Loader, self).load_data(**kwargs, )
        self.variables['DATETIME'] = self.variables['SC_DATETIME']
        self.variables['J_N'] = self.variables['J'][:, 0][:, np.newaxis]
        self.variables['J_E'] = self.variables['J'][:, 1][:, np.newaxis]
        self.variables['SC_GEO_ALT'] = np.ones_like(self.variables['SC_GEO_LAT']) * 110.



import datetime
import numpy as np
import pathlib
import re
import requests
import tqdm

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.config import prf


class Downloader(object):

    def __init__(
            self, dt_fr, dt_to,
            data_file_root_dir=None,
            data_product='grd', pole='N', user_name=None, direct_download=True, force_download=False):
        self.dt_fr = dt_fr
        self.dt_to = dt_to
        self.data_product = data_product
        self.user_name = user_name
        self.pole = pole
        self.force_download = force_download
        self.base_url = "https://ampere.jhuapl.edu/services"
        self.done = False
        self.data_file_root_dir = data_file_root_dir
        self.data_file_paths = []
        if self.data_file_root_dir is None:
            self.data_file_root_dir = prf.datahub_data_root_dir / 'JHUAPL' / 'AMPERE' / data_product.upper()

        if direct_download:
            self.download()

    def download(self):
        diff_days = dttool.get_diff_days(self.dt_fr, self.dt_to)
        dt_0 = dttool.get_start_of_the_day(self.dt_fr)
        for nd in range(diff_days+1):
            this_day = dt_0 + datetime.timedelta(days=nd)
            
            for hh in range(24):
             
                time_start = this_day + datetime.timedelta(hours=hh)
                time_extent = 3600 - 1
                time_end = time_start + datetime.timedelta(seconds=time_extent)
                if (time_start < self.dt_fr) or (time_end > self.dt_to):
                    continue
                params = {
                    'logon': self.user_name,
                    'start': time_start.strftime('%Y-%m-%dT%H:%M'),
                    'end': time_end.strftime('%Y-%m-%dT%H:%M'),
                    'pole': self.pole_str
                }
                url = self.base_url + '/data-' + self.data_product.lower() + '.php'
                file_name = '_'.join([
                    'AMPERE',
                    self.data_product.upper(),
                    time_start.strftime('%Y%m%dT%H%M'),
                    time_end.strftime('%Y%m%dT%H%M'),
                    self.pole,
                ]) + '.nc'
                file_dir = self.data_file_root_dir / dt_0.strftime("%Y%m%d")
                file_dir.mkdir(parents=True, exist_ok=True)
                fp = file_dir / file_name
                if fp.is_file():
                    if not self.force_download:
                        self.done = True
                        mylog.simpleinfo.info("The file {} exists in {}.".format(file_name, file_dir))
                        self.data_file_paths.append(fp)
                        continue
                mylog.simpleinfo.info("Downloading {} from the online database ...".format(file_name))
                status = self._download_by_requests_get(url, data_file_path=fp, params=params)
                if status == -1:
                    mylog.StreamLogger.warning(
                        f'Failed to download the AMPERE {self.data_product.upper()} data between' 
                        + f'{time_start.strftime("%Y%m%dT%H:%M")} and {time_end.strftime("%Y%m%dT%H:%M")}!')
                mylog.simpleinfo.info("Done. The file has been saved to {}".format(file_dir))
                self.data_file_paths.append(fp)
                self.done = True

    def _download_by_requests_get(self, url, data_file_path=None, params=None, stream=True, **kwargs):
        r = requests.get(url, params=params, stream=stream, **kwargs)
        if r.status_code != 200:
            return -1

        total_size_in_bytes = int(r.headers.get('content-length', 0))
        block_size = 1024  # 1 Kibibyte
        progress_bar = tqdm.tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
        
        with open(data_file_path, 'wb') as file:
            for data in r.iter_content(block_size):
                progress_bar.update(len(data))
                file.write(data)
        progress_bar.close()
        if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
            mylog.StreamLogger.error("Something wrong during the download!")
            data_file_path.unlink(missing_ok=True)
            return -1

        return 1

    @property
    def pole(self):
        return self._pole
    
    @pole.setter
    def pole(self, value):
        if value.upper() in ['N', 'NORTH']:
            self.pole_str = 'north'
        elif value.upper() in ['S', 'SOUTH']:
            self.pole_str = 'south'
        else:
            raise ValueError
        self._pole = value  


if __name__ == "__main__":
    dt_fr = datetime.datetime(2016, 3, 14, 1)
    dt_to = datetime.datetime(2016, 3, 15, 1)
    data_file_root_dir = pathlib.Path(__file__).parent.resolve()
    download = Downloader(dt_fr=dt_fr, dt_to=dt_to, data_file_root_dir=data_file_root_dir,
                          pole='N', user_name='leicai') 



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import h5py
import numpy as np
import pathlib
import re
import scipy.interpolate as si

from geospacelab.config import pref as prf

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.datahub.sources.madrigal.isr.eiscat as eiscat
import geospacelab.toolbox.utilities.pylogging as mylog


default_variable_names = [
    'DATETIME', 'DATETIME_1', 'DATETIME_2',
    'MAGIC_CONSTANT', 'r_SCangle', 'r_m0_1', 'r_m0_2'
    'AZ', 'EL', 'P_Tx', 'HEIGHT', 'RANGE',
    'n_e', 'T_i', 'T_e', 'nu_i', 'v_i_los', 'comp_mix', 'comp_O_p',
    'n_e_err', 'T_i_err', 'T_e_err', 'nu_i_err', 'v_i_los_err', 'comp_mix_err', 'comp_O_p_err',
    'STATUS', 'RESIDUAL'
]


class Loader:
    def __init__(self, file_path, file_type="eiscat-hdf5", gate_num=None):
        self.variables = {}
        self.metadata = {}
        self.file_path = file_path
        self.file_type = file_type
        self.gate_num = gate_num
        self.load_data()

    def load_data(self):
        if self.file_type == 'eiscat-hdf5':
            self.load_eiscat_hdf5()
        elif self.file_type == 'eiscat-mat':
            self.load_eiscat_mat()
        elif self.file_type == 'madrigal-hdf5':
            self.load_madrigal_hdf5()
        else:
            raise ValueError

    def load_eiscat_hdf5(self):
        """
        Load EISCAT hdf5 file.
        :param file_paths: a list of the file fullpaths for the hdf5 files
        :return:
            load_obj: instance of Loader
                .variables: queried variables
                .metadata: metadata
        """

        var_name_dict = {
            'MAGIC_CONSTANT': 'Magic_const',
            'r_SCangle': 'SCangle',
            'r_m0_2': 'm02',
            'r_m0_1': 'm01',
            'AZ': 'az',
            'EL': 'el',
            'P_Tx': 'Pt',
            'T_SYS_1': 'Tsys1',
            'T_SYS_2': 'Tsys2',
            'HEIGHT': 'h',
            'RANGE': 'range',
            'n_e': 'Ne',
            'T_i': 'Ti',
            'T_r': 'Tr',
            'nu_i': 'Collf',
            'v_i_los': 'Vi',
            'comp_mix': 'pm',
            'comp_O_p': 'po+',
            'n_e_var': 'var_Ne',
            'T_i_var': 'var_Ti',
            'T_r_var': 'var_Tr',
            'nu_i_var': 'var_Collf',
            'v_i_los_var': 'var_Vi',
            'comp_mix_var': 'var_pm',
            'comp_O_p_var': 'var_po+',
            'STATUS': 'status',
            'RESIDUAL': 'res1'
        }

        site_info = {
            'T': 'UHF',
            'V': 'VHF',
            'K': 'KIR',
            'S': 'SOD',
            'L': 'ESR'
        }

        vars = {}
        metadata = {}

        with h5py.File(self.file_path, 'r') as fh5:
            var_info_list = eiscat.list_eiscat_hdf5_variables(fh5)
            h5_data = fh5['data']
            h5_metadata = fh5['metadata']
            ind_nrec = var_info_list['name'].index('nrec')
            nrec_group = var_info_list['group'][ind_nrec]
            nrec_group_ind = var_info_list['index'][ind_nrec]
            nrec = h5_data[nrec_group][nrec_group_ind]
            num_row = h5_data['utime'][0].shape[0]
            for var_name, var_name_h5 in var_name_dict.items():
                try:
                    ind_v = var_info_list['name'].index(var_name_h5)
                except ValueError:
                    mylog.StreamLogger.warning(f"'{var_name_h5}' is not in the hdf5 file!")
                    vars[var_name] = None
                    continue
                var_group = var_info_list['group'][ind_v]
                var_ind = var_info_list['index'][ind_v]
                var = h5_data[var_group][var_ind]
                if var_group == 'par0d':
                    vars[var_name] = var[0]
                elif var_group == 'par1d':
                    var = var.reshape(num_row, 1)
                    vars[var_name] = var
                elif var_group == 'par2d':
                    if nrec_group == 'par0d':
                        num_col = int(var.shape[0] / num_row)
                        if num_col != nrec[0]:
                            print("Note: the number of range gates doesn't match nrec!")
                        var = var.reshape(num_row, num_col)
                        if self.gate_num is None:
                            num_gates=self.gate_num = num_col
                        else:
                            num_gates = self.gate_num
                        var_array = np.empty((num_row, num_gates))
                        var_array[::] = np.nan
                        for i in range(num_row):
                            var_array[i, 0:num_col] = var[i, :]
                        var = var_array
                    elif nrec_group == 'par1d':
                        if self.gate_num is None:
                            self.gate_num = num_gates = int(np.max(nrec))
                        else:
                            num_gates = self.gate_num
                        var_array = np.empty((num_row, num_gates))
                        var_array[:, :] = np.nan
                        rec_ind_1 = 0
                        for i in range(num_row):
                            rec_ind_2 = int(rec_ind_1 + nrec[i])
                            var_array[i, :int(nrec[i])] = var[rec_ind_1:rec_ind_2]
                            rec_ind_1 = rec_ind_2
                        var = var_array
                    vars[var_name] = var

            # unix time to datetime
            utime1 = h5_data['utime'][0]
            dt1 = dttool.convert_unix_time_to_datetime_cftime(utime1)
            var = dt1.reshape(num_row, 1)
            var_name = 'DATETIME_1'
            vars[var_name] = var

            utime2 = h5_data['utime'][1]
            dt2 = dttool.convert_unix_time_to_datetime_cftime(utime2)
            var = dt2.reshape(num_row, 1)
            var_name = 'DATETIME_2'
            vars[var_name] = var

            vars['P_Tx'] = vars['P_Tx'] / 1e3   # in kW

            metadata['r_XMITloc'] = [h5_data['par0d'][2][0], h5_data['par0d'][3][0], h5_data['par0d'][4][0]]
            metadata['r_RECloc'] = [h5_data['par0d'][5][0], h5_data['par0d'][6][0], h5_data['par0d'][7][0]]
            metadata['site_name'] = site_info[h5_metadata['names'][1][1].decode('UTF-8').strip()]
            metadata['pulse_code'] = h5_metadata['names'][0][1].decode('UTF-8').strip()
            metadata['antenna'] = h5_metadata['names'][2][1].decode('UTF-8').strip()
            metadata['GUISDAP_version'] = h5_metadata['software']['GUISDAP_ver'][0, 0].decode('UTF-8').strip()
            metadata['rawdata_path'] = ''
            metadata['scan_mode'] = ''
            metadata['affiliation'] = ''
            try:
                title = h5_metadata['schemes']['DataCite']['Title'][0][0].decode('UTF-8').strip()
                rc = re.compile(r'\d{4}-\d{2}-\d{2}_([\S]+)@')
                exp_id = rc.findall(title)[0]
                metadata['modulation'] = exp_id.split('_')[-1]
                metadata['pulse_code'] = exp_id.replace('_'+metadata['modulation'], '')
            except KeyError:
                mylog.StreamLogger.warning("'Title is not listed in the metadata!'")

            try:
                # 'gfd' not in list before 2001?
                metadata['rawdata_path'] = h5_metadata['software']['gfd']['data_path'][0, 0].decode('UTF-8').strip()
                metadata['scan_mode'] = metadata['rawdata_path'].split('_')[1]
                metadata['affiliation'] = metadata['rawdata_path'].split('@')[0].split('_')[-1]
            except KeyError:
                mylog.StreamLogger.warning(
                    "'gfd' is not listed in the metadata! Affect 'rawdata_path', 'scan_mode', and 'affiliation'."
                )

        vars_add = {}
        for var_name, value in vars.items():
            if '_var' in var_name:
                var_name_err = var_name.replace('_var', '_err')
                vars_add[var_name_err] = np.sqrt(vars[var_name])
        vars.update(vars_add)

        vars['DATETIME'] = vars['DATETIME_1'] + (vars['DATETIME_2'] - vars['DATETIME_1'])/2
        vars['T_e'] = vars['T_i'] * vars['T_r']
        vars['T_e_err'] = vars['T_e'] * np.sqrt((vars['T_i_err']/vars['T_i'])**2
                                                + (vars['T_r_err']/vars['T_r'])**2)
        vars['AZ'] = vars['AZ'] % 360.
        # check height and range
        vars['HEIGHT'] = vars['HEIGHT'] / 1000.
        vars['RANGE'] = vars['RANGE'] / 1000.

        inds_nan = np.where(np.isnan(vars['HEIGHT']))
        if list(inds_nan):
            m, n = vars['HEIGHT'].shape
            for i in range(m):
                yy = vars['HEIGHT'][i, :].flatten()
                iii = np.where(~np.isfinite(yy))
                if not list(iii):
                    continue
                iii = np.where(np.isfinite(yy))[0]
                xx = np.arange(0, n)
                f = si.interp1d(xx[iii], yy[iii], kind='linear', bounds_error=False, fill_value='extrapolate')
                yy_new = f(xx)
                vars['HEIGHT'][i, :] = yy_new

                yy = vars['RANGE'][i, :].flatten()
                iii = np.where(~np.isfinite(yy))
                if not list(iii):
                    continue
                iii = np.where(np.isfinite(yy))[0]
                xx = np.arange(0, n)
                f = si.interp1d(xx[iii], yy[iii], kind='linear', bounds_error=False, fill_value='extrapolate')
                yy_new = f(xx)
                vars['RANGE'][i, :] = yy_new

        if np.isscalar(vars['AZ']):
            az = np.empty((vars['DATETIME_1'].shape[0], 1))
            az[:, :] = vars['AZ']
            vars['AZ'] = az
        if np.isscalar(vars['EL']):
            el = np.empty((vars['DATETIME_1'].shape[0], 1))
            el[:, :] = vars['EL']
            vars['EL'] = el
        vars['AZ'] = np.mod(vars['AZ'], 360)
        vars['EL'] = np.mod(vars['EL'], 360)

        self.variables = vars
        self.metadata = metadata

    def load_eiscat_mat(self):

        raise NotImplemented

    def load_madrigal_hdf5(self):
        from scipy.signal import argrelmin, argrelmax

        var_name_dict = {
            # 'MAGIC_CONSTANT': 'Magic_const',
            'r_SCangle': 'HSA',
            # 'r_m0_2': 'm02',
            # 'r_m0_1': 'm01',
            'AZ': 'AZM',
            'EL': 'ELM',
            'P_Tx': 'POWER',
            'T_SYS_1': 'SYSTMP',
            'T_SYS_2': 'SYSTMP',
            'HEIGHT': 'GDALT',
            'RANGE': 'RANGE',
            'n_e': 'NE',
            'T_i': 'TI',
            'T_r': 'TR',
            'nu_i': 'CO',
            'v_i_los': 'VO',
            'comp_mix': 'PM',
            'comp_O_p': 'PO+',
            'n_e_err': 'DNE',
            'T_i_err': 'DTI',
            'T_r_err': 'DTR',
            'nu_i_err': 'DCO',
            'v_i_los_err': 'DVO',
            'comp_mix_err': 'DPM',
            'comp_O_p_err': 'DPO+',
            'STATUS': 'GFIT',
            'RESIDUAL': 'CHISQ'
        }
        with h5py.File(self.file_path, 'r') as fh5:

            # load metadata
            metadata = {}

            exp_params = fh5['Metadata']['Experiment Parameters'][:]
            exp_params = list(zip(*tuple(exp_params)))
            fn_id = exp_params[0].index(b'Cedar file name')
            fn = exp_params[1][fn_id].decode('UTF-8')
            rc = re.compile(r'^MAD[\w]+_[\d]{4}-[\d]{2}-[\d]{2}_([\w.]+)@(\w+)\.hdf5')
            rm = rc.findall(fn)[0]
            pattern_1 = rm[0]
            pattern_2 = rm[1]
            metadata['modulation'] = pattern_1.split('_')[-1]
            metadata['pulse_code'] = pattern_1.replace('_' + metadata['modulation'], '')

            antenna = pattern_2
            if 'uhf' in antenna:
                sitename = 'UHF'
            elif 'vhf' in antenna:
                sitename = 'VHF'
            elif 'sod' in antenna:
                sitename = 'SOD'
            elif 'kir' in antenna:
                sitename = 'KIR'
            elif '32' in antenna or '42' in antenna:
                sitename = 'ESR'
            else:
                print(antenna)
                raise AttributeError
            metadata['site_name'] = sitename

            lat_id = exp_params[0].index(b'instrument latitude')
            lat = float(exp_params[1][lat_id].decode('UTF-8'))
            lon_id = exp_params[0].index(b'instrument longitude')
            lon = float(exp_params[1][lon_id].decode('UTF-8'))
            alt_id = exp_params[0].index(b'instrument altitude')
            alt = float(exp_params[1][alt_id].decode('UTF-8')) * 1000

            metadata['r_RECloc'] = [lat, lon, alt]
            if sitename in ['ESR', 'UHF', 'VHF', 'TRO']:
                metadata['r_XMITloc'] = [lat, lon, alt]
            else:
                metadata['r_XMITloc'] = [69.583, 19.21, 30]
            metadata['antenna'] = antenna
            metadata['GUISDAP_version'] = ''
            metadata['rawdata_path'] = ''
            metadata['scan_mode'] = ''
            metadata['affiliation'] = ''
            metadata['rawdata_path'] = ''
            metadata['scan_mode'] = ''
            metadata['affiliation'] = ''

            # load data
            data = fh5['Data']['Table Layout'][:]
            data = list(zip(*tuple(data)))
            data_parameters = list(zip(*tuple(fh5['Metadata']['Data Parameters'][:])))
            var_names_h5 = [vn.decode('UTF-8') for vn in data_parameters[0]]
            nvar_h5 = len(var_names_h5)
            ran_id = var_names_h5.index('RANGE')
            ran = data[ran_id]
            inds_ran_min = argrelmin(np.array(ran))[0]
            inds_ran_min = np.append(inds_ran_min, 0)
            inds_ran_min.sort()

            inds_ran_max = argrelmax(np.array(ran))[0]
            inds_ran_max = np.append(inds_ran_max, len(ran)-1)
            inds_ran_max.sort()

            if self.gate_num is None:
                self.gate_num = ngates_max = np.max(np.diff(inds_ran_max))
            else:
                ngates_max = self.gate_num
            num_row = inds_ran_min.shape[0]
            data_array = np.empty((nvar_h5, num_row, ngates_max))
            data_array[::] = np.nan
            for ip in range(nvar_h5):
                var_tmp = np.array(data[ip])
                for i in range(inds_ran_min.shape[0]):
                    ind1 = inds_ran_min[i]
                    ind2 = inds_ran_max[i]
                    data_array[ip, i, 0: ind2-ind1+1] = var_tmp[ind1: ind2+1]

            vars_h5 = {}
            for ip in range(nvar_h5):

                var = data_array[ip]
                test_unique = np.unique(var[0][~np.isnan(var[0])])
                if test_unique.shape[0] == 1 and not re.match(r'^D\w+', var_names_h5[ip]) and var_names_h5[ip] not in ['GFIT']:
                    # print(var_names_h5[ip])
                    var = np.array([list(set(var[i][~np.isnan(var[i])])) for i in range(num_row)]).reshape((num_row, 1))
                vars_h5[var_names_h5[ip]] = var

            vars = {}
            for var_name, var_name_h5 in var_name_dict.items():
                try:
                    vars[var_name] = vars_h5[var_name_h5]
                except KeyError:
                    vars[var_name] = None

            num_row = inds_ran_min.shape[0]
            # unix time to datetime
            utime1 = vars_h5['UT1_UNIX'].flatten()
            dt1 = dttool.convert_unix_time_to_datetime_cftime(utime1)
            var = dt1.reshape(num_row, 1)
            var_name = 'DATETIME_1'
            vars[var_name] = var

            utime2 = vars_h5['UT2_UNIX'].flatten()
            dt2 = dttool.convert_unix_time_to_datetime_cftime(utime2)
            var = dt2.reshape(num_row, 1)
            var_name = 'DATETIME_2'
            vars[var_name] = var

            vars['DATETIME'] = vars['DATETIME_1'] + (vars['DATETIME_2'] - vars['DATETIME_1']) / 2
            vars['T_e'] = vars['T_i'] * vars['T_r']
            vars['T_e_err'] = vars['T_e'] * np.sqrt((vars['T_i_err'] / vars['T_i']) ** 2
                                                    + (vars['T_r_err'] / vars['T_r']) ** 2)
            vars['AZ'] = vars['AZ'] % 360.

            inds_nan = np.where(np.isnan(vars['HEIGHT']))
            if list(inds_nan):
                m, n = vars['HEIGHT'].shape
                for i in range(m):
                    yy = vars['HEIGHT'][i, :].flatten()
                    iii = np.where(~np.isfinite(yy))
                    if not list(iii):
                        continue
                    iii = np.where(np.isfinite(yy))[0]
                    xx = np.arange(0, n)
                    f = si.interp1d(xx[iii], yy[iii], kind='linear', bounds_error=False, fill_value='extrapolate')
                    yy_new = f(xx)
                    vars['HEIGHT'][i, :] = yy_new

                    yy = vars['RANGE'][i, :].flatten()
                    iii = np.where(~np.isfinite(yy))
                    if not list(iii):
                        continue
                    iii = np.where(np.isfinite(yy))[0]
                    xx = np.arange(0, n)
                    f = si.interp1d(xx[iii], yy[iii], kind='linear', bounds_error=False, fill_value='extrapolate')
                    yy_new = f(xx)
                    vars['RANGE'][i, :] = yy_new


            self.variables = vars
            self.metadata = metadata
        # raise NotImplemented


if __name__ == "__main__":
    dir_root = prf.datahub_data_root_dir

    # fp = pathlib.Path('examples') / "EISCAT_2005-09-01_steffe_64@32m.hdf5"
    fp = pathlib.Path('examples') / "MAD6400_2021-03-10_beata_ant@uhfa.hdf5"
    Loader(file_path=fp, file_type = 'madrigal-hdf5')



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import numpy as np


def convert_local_az_el_range_to_geo(lat_0, lon_0, height_0, az, el, beam_range=None, radians = False):
    if radians:
        rd = 1.
    else:
        rd = np.pi / 180.
    lat_0 = lat_0 * rd
    lon_0 = lon_0 * rd
    az = az * rd
    el = el * rd





import datetime
import pathlib
import copy

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.config import prf
from geospacelab.datahub.sources.cdaweb.downloader import Downloader as DownloaderBase


class Downloader(DownloaderBase):
    
    def __init__(
            self,
            dt_fr=None, dt_to=None,
            sat_id=None,
            orbit_id=None,
            direct_download=True,
            force_download=False,
            data_file_root_dir = None,
            dry_run=False,
        ):
        product = 'EDR_AUR'
        if data_file_root_dir is None:
            data_file_root_dir = prf.datahub_data_root_dir / 'CDAWeb' / 'DMSP' / 'SSUSI' / product
        self.sat_id = sat_id
        self.orbit_id = orbit_id
        self.source_subdirs = ['dmsp', 'dmsp'+self.sat_id.lower(), 'ssusi', 'data', 'edr-aurora']

        super().__init__(
            dt_fr, dt_to,
            data_file_root_dir=data_file_root_dir,
            direct_download=direct_download,force_download=force_download,dry_run=dry_run
        )


    def search_from_http(self, file_name_patterns=None, allow_multiple_files=True):

        dt_fr_1 = self.dt_fr - datetime.timedelta(hours=3)
        dt_to_1 = self.dt_to + datetime.timedelta(hours=3)
        diff_days = dttool.get_diff_days(dt_fr_1, dt_to_1)
        dt0 = dttool.get_start_of_the_day(dt_fr_1)
        source_file_paths = []
        for nd in range(diff_days + 1):
            this_day = dt0 + datetime.timedelta(days=nd)
            doy = dttool.get_doy(this_day)
            sdoy = '{:03d}'.format(doy)
            subdirs = copy.deepcopy(self.source_subdirs)
            subdirs.extend(
                [str(this_day.year), sdoy]
            )

            if self.orbit_id is None:
                file_name_patterns = [
                    'dmsp' + self.sat_id.lower(),
                    'ssusi',
                    'edr-aurora',
                    this_day.strftime("%Y") + sdoy + 'T',
                    '.nc'
                ]
            else:
                file_name_patterns = [
                    'dmsp' + self.sat_id.lower(),
                    'ssusi',
                    'edr-aurora',
                    'REV',
                    self.orbit_id,
                    '.nc'
                ]
            paths = super().search_from_http(subdirs=subdirs, file_name_patterns=file_name_patterns)
            source_file_paths.extend(paths)
        return source_file_paths

    def save_file_from_http(self, url, file_dir=None, file_name=None):

        sy = url.split('/')[-3]
        sdoy = url.split('/')[-2]
        year = int(sy)
        this_day = dttool.convert_doy_to_datetime(year, int(sdoy))
        if file_dir is None:
            file_dir = self.data_file_root_dir /  self.sat_id.upper() / sy / this_day.strftime("%Y%m%d")
        super().save_file_from_http(url, file_dir=file_dir)



if __name__ == "__main__":
    downloader = Downloader(
        dt_fr = datetime.datetime(2011, 1, 6),
        dt_to = datetime.datetime(2011, 1, 6, 12),
        sat_id='F17',
        orbit_id='21523',
        dry_run=False,
    )


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import re

import netCDF4 as nc
import numpy as np
import cftime
import datetime

import geospacelab.toolbox.utilities.pydatetime as dttool


class Loader:
    """
    :param file_path: the file's full path
    :type file_path: pathlib.Path object
    :param file_type: the specific file type for the file being loaded. Options: ['TEC-MAT'], 'TEC-LOS', 'TEC-sites')
    :type file_type: str
    :param load_data: True, load without calling the method "load_data" separately.
    :type load_data: bool
    """
    def __init__(self, file_path, file_type='ATEC', load_data=True):
        self.file_path: pathlib.Path = file_path
        self.file_type = file_type
        self.variables = {}
        self.done = False
        if load_data:
            self.load()

    def load(self):
        if self.file_type == "ATEC":
            self.load_tec_map()
        else:
            raise NotImplemented

    def load_tec_map(self):
        fnc = nc.Dataset(self.file_path)
        file_name = self.file_path.stem
        rc = re.compile(r"(\d{10})")
        tstr = rc.search(file_name).groups()[0]
        dt_0 = dttool.get_start_of_the_day(datetime.datetime.strptime(tstr, "%Y%m%d%H"))

        variables = {}

        glat = np.array(fnc.variables['lat'])
        glon = np.array(fnc.variables['lon'])
        variables['GEO_LAT'] = np.tile(glat.reshape(glat.size, 1), (1, glon.size))
        variables['GEO_LON'] = np.tile(glon.reshape(1, glon.size), (glat.size, 1))
        tec_map = np.array(fnc.variables['atec']).transpose((2, 0, 1))
        tec_map = np.where(tec_map >= 900., np.nan, tec_map)
        variables['TEC_MAP'] = tec_map
        time = np.array(fnc.variables['time'])
        dts = np.array(
            [dt_0 + datetime.timedelta(seconds=t) for t in time]
        )
        variables['DATETIME'] = dts[:, np.newaxis]

        self.variables = variables
        self.done = True
        fnc.close()


nc_variable_name_dict = {
    'GEO_LAT': 'lat',
    'GEO_LON': 'lon',
    'TEC_MAP': 'atec',
}


if __name__ == "__main__":
    import pathlib
    fp = pathlib.Path("/home/lei/Downloads/2023010702_atec.nc")
    Loader(file_path=fp)
    # import pathlib
    # import geospacelab.datahub.sources.madrigal as madrigal
    # file_path = pathlib.Path("/Users/lcai/Downloads/gps200102g.002.hdf5")
    # madrigal.utilities.show_hdf5_structure(filename=file_path.name, filepath= file_path.parent.resolve())
    # madrigal.utilities.show_hdf5_metadata(filename=file_path.name, filepath=file_path.parent.resolve())
    # with h5py.File(file_path, 'r') as fh5:
    #
    #     print(fh5['Data']['Array Layout']['2D Parameters']['tec'][:])  # shape (180, 360, 288)
    #     print(fh5['Data']['Array Layout']['timestamps'][:])  # shape (288) unix time
    #     print(fh5['Data']['Array Layout']['glon'][:])  # gdlat -90. - 89. glon -180. 79.


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import datetime
import numpy as np

from geospacelab.datahub.sources.madrigal.isr.millstonehill.gridded.loader import Loader as default_Loader
from geospacelab.datahub.sources.madrigal.isr.millstonehill.downloader import Downloader as default_Downloader
import geospacelab.datahub.sources.madrigal.isr.millstonehill.gridded.variable_config as var_config
from geospacelab.datahub.sources.madrigal import madrigal_database
from geospacelab.config import prf
from geospacelab import datahub
from geospacelab.datahub import SiteModel, DatabaseModel, FacilityModel
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog

default_dataset_attrs = {
    'kind': 'sourced',
    'database': madrigal_database,
    'facility': 'MillstoneHillISR',
    'exp_name_pattern': '',
    'exp_check': False,
    'data_file_type': 'gridded',
    'data_file_ext': 'hdf5',
    'data_root_dir': prf.datahub_data_root_dir / 'Madrigal' / 'MillstoneHill_ISR',
    'allow_download': True,
    'status_control': False,
    'residual_control': False,
    'beam_location': True,
    'data_search_recursive': True,
    'label_fields': ['database', 'facility', 'site', 'data_file_type'],
}

default_variable_names = [
    'DATETIME', 'HEIGHT', 'n_pp_log', 'n_pp_log_err', 'n_e', 'n_e_err', 'T_i', 'T_i_err', 'T_r',
    'T_e', 'T_e_err', 'v_i_UP', 'v_i_UP_err',
    'T_r_err', 'comp_mix', 'comp_mix_err', 'coef_T_i_T_r', 'coef_T_i_T_r',
    'GEO_LAT', 'GEO_LON', 'TEC', 'n_e_max', 'h_max', 'GEO_ALT',
]

# default_data_search_recursive = True

default_attrs_required = []


class Dataset(datahub.DatasetSourced):
    def __init__(self, **kwargs):
        kwargs = basic.dict_set_default(kwargs, **default_dataset_attrs)

        super().__init__(**kwargs)

        self.database = kwargs.pop('database', '')
        self.facility = kwargs.pop('facility', '')
        self.site = kwargs.pop('site', MillstoneHillSite('MillstoneHill'))
        self.experiment = kwargs.pop('experiment', '')
        self.exp_name_pattern = kwargs.pop('exp_name_pattern', '')
        self.exp_check = kwargs.pop('exp_check', False)
        self.data_file_type = kwargs.pop('data_file_type', '')
        self.affiliation = kwargs.pop('affiliation', '')
        self.allow_download = kwargs.pop('allow_download', True)
        self.beam_location = kwargs.pop('beam_location', True)
        self.metadata = None

        allow_load = kwargs.pop('allow_load', True)
        self.status_control = kwargs.pop('status_control', False)
        self.residual_control = kwargs.pop('residual_control', False)

        # self.config(**kwargs)

        if self.loader is None:
            self.loader = default_Loader
        if self.downloader is None:
            self.downloader = default_Downloader

        self._set_default_variables(
            default_variable_names,
            configured_variables=var_config.configured_variables
        )

        self._validate_attrs()

        if allow_load:
            self.load_data()

    def _validate_attrs(self):
        for attr_name in default_attrs_required:
            attr = getattr(self, attr_name)
            if not str(attr):
                mylog.StreamLogger.warning("The parameter {} is required before loading data!".format(attr_name))

        if self.exp_check:
            self.download_data()

    def label(self, **kwargs):
        label = super().label()
        return label

    def load_data(self, **kwargs):
        self.check_data_files(**kwargs)

        for file_path in self.data_file_paths:
            load_obj = self.loader(file_path, )

            for var_name in self._variables.keys():
                self._variables[var_name].join(load_obj.variables[var_name])
            self.metadata = load_obj.metadata

        if self.time_clip:
            self.time_filter_by_range()

    def search_data_files(self, recursive=True, **kwargs):
        dt_fr = self.dt_fr
        dt_to = self.dt_to
        diff_days = dttool.get_diff_days(dt_fr, dt_to)
        day0 = dttool.get_start_of_the_day(dt_fr)
        for i in range(diff_days + 1):
            thisday = day0 + datetime.timedelta(days=i)
            initial_file_dir = self.data_root_dir / thisday.strftime('%Y') / thisday.strftime('%Y%m%d')

            file_patterns = ['MillstoneHill', self.data_file_type.replace(' ', '_'), thisday.strftime('%Y%m%d')]

            # remove empty str
            file_patterns = [pattern for pattern in file_patterns if str(pattern)]

            search_pattern = '*' + '*'.join(file_patterns) + '*'

            if str(self.exp_name_pattern):
                search_pattern = '*' + self.exp_name_pattern.replace(' ', '-') + '*/' + search_pattern
                recursive = False

            done = super().search_data_files(
                initial_file_dir=initial_file_dir, search_pattern=search_pattern, recursive=recursive)
            # Validate file paths

            if not done and self.allow_download:
                done = self.download_data()
                if done:
                    done = super().search_data_files(
                        initial_file_dir=initial_file_dir, search_pattern=search_pattern)
                else:
                    print('The requested experiment does not exist in the online database!')

            if len(done) > 1:
                if str(self.exp_name_pattern):
                    mylog.StreamLogger.error(
                        "Multiple data files detected! Check the files:")
                else:
                    mylog.StreamLogger.error(
                        "Multiple data files detected!" +
                        "Specify the experiment name by the keyword 'exp_name_pattern' if possible.")
                for fp in done:
                    mylog.simpleinfo.info(fp)
                raise KeyError
        return done

    def download_data(self):
        download_obj = self.downloader(
            dt_fr=self.dt_fr, dt_to=self.dt_to,
            data_file_root_dir=self.data_root_dir,
            file_type=self.data_file_type,
            exp_name_pattern=self.exp_name_pattern)
        return download_obj.done

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        if isinstance(value, str):
            self._database = DatabaseModel(value)
        elif issubclass(value.__class__, DatabaseModel):
            self._database = value
        else:
            raise TypeError

    @property
    def facility(self):
        return self._facility

    @facility.setter
    def facility(self, value):
        if isinstance(value, str):
            self._facility = FacilityModel(value)
        elif issubclass(value.__class__, FacilityModel):
            self._facility = value
        else:
            raise TypeError

    @property
    def site(self):
        return self._site

    @site.setter
    def site(self, value):
        if isinstance(value, str):
            self._site = MillstoneHillSite(value)
        elif issubclass(value.__class__, SiteModel):
            self._site = value
        else:
            raise TypeError


class MillstoneHillSite(SiteModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj

    def __init__(self, str_in, **kwargs):
        self.name = 'MillstoneHill'
        self.location = {
            'GEO_LAT': 42.619,
            'GEO_LON': 288.51,
            'GEO_ALT': 0.146
        }















# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

import netCDF4
import datetime
import numpy as np
import geospacelab.toolbox.utilities.pydatetime as dttool


class Loader(object):

    def __init__(self, file_path, file_type='edr-aur', pole='N'):

        self.variables = {}
        self.metadata = {}
        self.file_path = file_path
        self.file_type = file_type
        self.pole = pole

        self.load_data()

    def load_data(self):
        dataset = netCDF4.Dataset(self.file_path)
        variables = {}
        metadata = {}

        if self.pole == 'N':
            pole = self.pole
            pole_str = 'NORTH'
        elif self.pole == 'S':
            pole = self.pole
            pole_str = 'SOUTH'
        else:
            raise ValueError
        # Time and Position
        # sectime = int(np.array(dataset.variables['TIME']).flatten()[0])
        # doy = int(np.array(dataset.variables['DOY']).flatten()[0])
        # year = int(np.array(dataset.variables['YEAR']).flatten()[0])
        # dt0 = dttool.convert_doy_to_datetime(year, doy)
        starting_time = datetime.datetime.strptime(dataset.STARTING_TIME, "%Y%j%H%M%S")
        variables['STARTING_TIME'] = starting_time
        stopping_time = datetime.datetime.strptime(dataset.STOPPING_TIME, "%Y%j%H%M%S")
        variables['STOPPING_TIME'] = stopping_time
        dt0 = dttool.get_start_of_the_day(starting_time)

        variables['SC_LAT'] = np.array(dataset.variables['LATITUDE'])
        variables['SC_LON'] = np.array(dataset.variables['LONGITUDE'])
        variables['SC_ALT'] = np.array(dataset.variables['ALTITUDE'])

        variables['GRID_MLAT'] = np.array(dataset.variables['LATITUDE_GEOMAGNETIC_GRID_MAP'])
        variables['GRID_MLON'] = np.array(
            dataset.variables['LONGITUDE_GEOMAGNETIC_' + pole_str + '_GRID_MAP'])
        variables['GRID_MLT'] = np.array(dataset.variables['MLT_GRID_MAP'])
        if self.pole == 'S':
            variables['GRID_MLAT'] = - variables['GRID_MLAT']
        variables['GRID_UT'] = np.array(dataset.variables['UT_' + pole])
        lat = np.array(variables['GRID_MLAT'])
        ut = np.array(variables['GRID_UT'])
        lat = np.where(ut == 0, np.nan, lat)
        if self.pole == 'N':
            ind_mid_t = np.where(lat == np.nanmax(lat.flatten()))
        else:
            ind_mid_t = np.where(lat == np.nanmin(lat.flatten()))
        sectime0 = variables['GRID_UT'][ind_mid_t][0] * 3600

        diff_days = dttool.get_diff_days(starting_time, stopping_time)
        if diff_days > 0 and sectime0 < 0.5 * 86400.:
            dt = dt0 + datetime.timedelta(seconds=int(sectime0 + 86400))
        else:
            dt = dt0 + datetime.timedelta(seconds=int(sectime0))
        variables['DATETIME'] = dt

        invalid_ut_inds = np.where(ut == 0)
        # Auroral map, #colors: 0: '1216', 1: '1304', 2: '1356', 3: 'LBHS', 4: 'LBHL'.
        variables['EMISSION_SPECTRA'] = ['1216', '1304', '1356', 'LBHS', 'LBHL']
        disk_aur = np.array(dataset.variables['DISK_RADIANCEDATA_INTENSITY_' + pole_str])
        # disk_aur[:, invalid_ut_inds] = np.nan
        disk_aur[disk_aur <= 0] = 0.1
        variables['GRID_AUR_1216'] = disk_aur[0, ::]
        variables['GRID_AUR_1216'][invalid_ut_inds] = np.nan
        variables['GRID_AUR_1304'] = disk_aur[1, ::]
        variables['GRID_AUR_1304'][invalid_ut_inds] = np.nan
        variables['GRID_AUR_1356'] = disk_aur[2, ::]
        variables['GRID_AUR_1356'][invalid_ut_inds] = np.nan
        variables['GRID_AUR_LBHS'] = disk_aur[3, ::]
        variables['GRID_AUR_LBHS'][invalid_ut_inds] = np.nan
        variables['GRID_AUR_LBHL'] = disk_aur[4, ::]
        variables['GRID_AUR_LBHL'][invalid_ut_inds] = np.nan

        # Auroral oval boundary
        variables['AOB_EQ_MLAT'] = np.array(dataset.variables[pole_str + '_GEOMAGNETIC_LATITUDE'])
        variables['AOB_EQ_MLON'] = np.array(dataset.variables[pole_str + '_GEOMAGNETIC_LONGITUDE'])
        variables['AOB_EQ_MLT'] = np.array(dataset.variables[pole_str + '_MAGNETIC_LOCAL_TIME'])

        variables['AOB_PL_MLAT'] = np.array(dataset.variables[pole_str + '_POLAR_GEOMAGNETIC_LATITUDE'])
        variables['AOB_PL_MLON'] = np.array(dataset.variables[pole_str + '_POLAR_GEOMAGNETIC_LONGITUDE'])
        variables['AOB_PL_MLT'] = np.array(dataset.variables[pole_str + '_POLAR_MAGNETIC_LOCAL_TIME'])

        variables['MAOB_EQ_MLAT'] = np.array(dataset.variables['MODEL_' + pole_str + '_GEOMAGNETIC_LATITUDE'])
        variables['MAOB_EQ_MLON'] = np.array(dataset.variables['MODEL_' + pole_str + '_GEOMAGNETIC_LONGITUDE'])
        variables['MAOB_EQ_MLT'] = np.array(dataset.variables['MODEL_' + pole_str + '_MAGNETIC_LOCAL_TIME'])

        variables['MAOB_PL_MLAT'] = np.array(dataset.variables['MODEL_' + pole_str + '_POLAR_GEOMAGNETIC_LATITUDE'])
        variables['MAOB_PL_MLON'] = np.array(dataset.variables['MODEL_' + pole_str + '_POLAR_GEOMAGNETIC_LONGITUDE'])
        variables['MAOB_PL_MLT'] = np.array(dataset.variables['MODEL_' + pole_str + '_POLAR_MAGNETIC_LOCAL_TIME'])

        metadata.setdefault('ORBIT_ID', dataset.STARTING_ORBIT_NUMBER)
        dataset.close()

        self.variables = variables
        self.metadata = metadata



import re
import pathlib
import datetime
import numpy as np

import geospacelab.datahub.sources.supermag.supermag_api as smapi
from geospacelab.config import prf
import geospacelab.toolbox.utilities.pylogging as mylog


def list_available_sites(dt_fr=None, dt_to=None, extent=None, user_id=None):
    if user_id is None:
        user_id = prf.user_config['datahub']['supermag']['username']

    extent = (dt_to - dt_fr).total_seconds() if extent is None else extent
    (success, sites_available) = smapi.SuperMAGGetInventory(user_id, dt_fr, extent)
    if not success:
        raise ConnectionError

    all_site_info = load_site_info()

    all_sites = all_site_info['CODE']
    need_update = []

    site_info_new = {
        'CODE': [],
        'NAME': [],
        'GLAT': [],
        'GLON': [],
        'OPERATORS': [],
    }
    for s in sites_available:
        ii = np.where(all_sites == s)[0]
        if not list(ii):
            need_update.append(s)
            continue
        site_info_new['CODE'].append(all_site_info['CODE'][ii[0]])
        site_info_new['NAME'].append(all_site_info['NAME'][ii[0]])
        site_info_new['GLAT'].append(all_site_info['GLAT'][ii[0]])
        site_info_new['GLON'].append(all_site_info['GLON'][ii[0]])
        site_info_new['OPERATORS'].append(all_site_info['OPERATORS'][ii[0]])

    if list(need_update):
        mylog.StreamLogger.warning("Cannot find the following sites, update the SuperMAG station records is needed!")
        mylog.simpleinfo.info(need_update)

    for k, v in site_info_new:
        site_info_new[k] = np.array(v)

    return site_info_new


def get_supermag_sites(
        dt_fr=None, dt_to=None, extent=None, user_id=None,
        lon_1=None, lon_2=None,
        lat_1=None, lat_2=None,
):
    lon_1 = 0. if lon_1 is None else lon_1
    lon_2 = 360. if lon_2 is None else lon_2

    lat_1 = - 90. if lat_1 is None else lat_1
    lat_2 = 90. if lat_2 is None else lat_2

    site_info = list_available_sites(dt_fr=dt_fr, dt_to=dt_to, extent=extent, user_id=user_id)

    glons = site_info['GLON']
    glats = site_info['GLAT']
    if lon_1 < lon_2:
        inds_lon = np.where((glons >= lon_1) & (glons < lon_2))[0]
    else:
        inds_lon = np.where((glons >= lon_1) & (glons < 360.))[0]
        inds_lon = np.concatenate((inds_lon, np.where((glons >=0) & (glons < lon_2))[0]))
    inds_lat = np.where((glats >= lat_1) & (glats <= lat_2))[0]
    inds = np.intersect1d(inds_lat, inds_lon)
    if not list(inds):
        return None

    site_info_new = {}
    for k, v in site_info.items():
        site_info_new[k] = v[inds]

    return site_info_new


def load_site_info(file_path=None, ):
    if file_path is None:
        file_path = pathlib.Path("SuperMAG_stations.dat")

    site_info = {
        'CODE': [],
        'NAME': [],
        'GLAT': [],
        'GLON': [],
        'OPERATORS': [],
    }
    with open(file_path, 'r') as f:
        text = f.read()

        results = re.findall(
            r'^(\w+)\s+([-.\d]+)\s+([-.\d]+)\s+([-.\d]+)\s+([-.\d]+)\s+"([^"]*)"\s+(\d+)\s+(.+)',
            text,
            re.M
        )
        results = list(zip(*results))
        site_info['CODE'] = np.array(results[0])
        site_info['NAME'] = np.array(results[5])
        site_info['GLAT'] = np.array(results[2]).astype(np.float32)
        site_info['GLON'] = np.array(results[1]).astype(np.float32) % 360

        ops = results[7]
        ops = [[i.strip('"') for i in op.split('\t')] for op in ops]
        site_info['OPERATORS'] = np.array(ops)

    return site_info


if __name__ == "__main__":
    # load_site_info()
    dt_fr_1 = datetime.datetime(2016, 3, 14)
    list_available_sites(dt_fr=dt_fr_1, extent=3600, user_id='JouleCai')


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import numpy
import numpy as np
from scipy.interpolate import interp1d, griddata
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pylogging as mylog
import datetime


def numpy_array_join_vertical(arr1, arr2):

    ndim2 = arr2.ndim
    if ndim2 == 1:
        ashape = (1, arr2.shape[0])
        arr2 = arr2.reshape(ashape)

    if arr1 is None:
        arr1 = numpy.empty((0, arr2.shape[1]))
    else:
        ndim1 = arr1.ndim
        if ndim1 == 1:
            ashape = (1, arr1.shape[0])
            arr1 = arr1.reshape(ashape)
    newarr = numpy.vstack((arr1, arr2))
    return newarr


def numpy_array_join_horizontal(arr1, arr2):
    ndim2 = arr2.ndim
    if ndim2 == 1:
        ashape = (arr2.shape[0], 1)
        arr2 = arr2.reshape(ashape)

    if arr1 is None:
        arr1 = numpy.empty((arr2.shape[0], 0))
    else:
        ndim1 = arr1.ndim
        if ndim1 == 1:
            ashape = (arr1.shape[0], 1)
            arr1 = arr1.reshape(ashape)
    newarr = numpy.hstack((arr1, arr2))
    return newarr


def numpy_array_self_mask(data, conditions=None):
    # conditions should be a list
    if conditions is None:
        return data
    for condition in conditions:
        if isinstance(condition, (float, int, str)):
            data = numpy.ma.masked_where(data == condition, data)
        elif isinstance(condition, dict):
            for key, value in condition.items():
                if key == 'between':
                    data = numpy.ma.masked_where((data >= value[0]) & (data <= value[1]), data)
                elif key == 'out':
                    data = numpy.ma.masked_where((data < value[0]) | (data > value[1]), data)
                elif key == 'greater':
                    data = numpy.ma.masked_where(data > value, data)
                elif key == 'smaller':
                    data = numpy.ma.masked_where(data < value, data)
                elif key == 'eq':
                    if not isinstance(value, list):
                        value = [value]
                    for v in value:
                        data = numpy.ma.masked_where(data == value, data)
                elif key == 'neq':
                    data = numpy.ma.masked_where(data != value, data)
    return data


def data_resample(
        x=None, y=None, xtype=None, xres=None, xresscale=1.1,
        method='Null',  # Null - insert NaN, 'linear', 'cubic', ... (interpolation method)
        axis=0, forward=True, depth=0.
):

    x1 = x
    if xtype == 'datetime':
        # dt0 = datetime.datetime(1970, 1, 1)
        sectime, dt0 = dttool.convert_datetime_to_sectime(x1)
        x1 = sectime

    diff_x1 = numpy.diff(x1.flatten())
    if xres is None:
        xres = numpy.median(diff_x1)
    inds = numpy.where(diff_x1 > xres * xresscale)[0]

    if len(inds) == 0:
        return x, y
    if depth > 10:
        return x, y

    inds = [i+1 for i in inds]
    # for x
    value = []
    for ind in inds:
        res = xres
        if xtype == 'datetime':
            res = datetime.timedelta(seconds=xres)
        if forward:
            value.append(x[ind - 1] + res/10)
        else:
            value.append(x[ind] - res/10)
    xnew = numpy.insert(x, inds, value, axis=axis)

    # for y
    if method == 'Null':
        value = numpy.nan
        ynew = numpy.insert(y, inds, value, axis=axis)
    else:
        ifunc = interp1d(x1.flatten(), y, kind=method, axis=axis)
        x_p = xnew
        if xtype=='datetime':
            x_p, dt0 = dttool.convert_datetime_to_sectime(xnew)
        ynew = ifunc(x_p.flatten())

    if method == 'Null':
        xnew, ynew = data_resample(
        x=xnew, y=ynew, xtype=xtype, xres=xres, xresscale=xresscale,
        method='Null',  # Null - insert NaN, 'linear', 'cubic', ... (interpolation method)
        axis=axis, forward=False, depth=depth+1
)

    return xnew, ynew

# def resample_2d(
#         x_data, y_data, z_data,
#         x_data_type=None,
#         x_data_res=None,
#         x_data_res_scale=1.,
#         x_grid_res=None,
#         x_grid_res_scale=1,
#         x_grid_min=None,
#         x_grid_max=None,
#         along_x_interp=True,
#         along_x_interp_method='nearest',
#         along_x_binning=False,
#         y_data_res=None,
#         y_data_res_scale=1.,
#         y_grid_res=None,
#         y_grid_res_scale=1.,
#         y_grid_min=None,
#         y_grid_max=None,
#         along_y_interp=True,
#         along_y_interp_method='nearest',
#         along_y_binning=False,
# ):
#     if along_x_interp==True and along_x_binning==True:
#         mylog.StreamLogger.error('The keywords "along_x_interp" and "along_x_binning" cannot be True at the same time!')
#         raise ValueError
#     if along_y_interp==True and along_y_binning==True:
#         mylog.StreamLogger.error('The keywords "along_y_interp" and "along_y_binning" cannot be True at the same time!')
#         raise ValueError
#
#     num_x, num_y = z_data.shape
#     if x_data_type == 'datetime':
#         dt0 = dttool.get_start_of_the_day(numpy.nanmin(x_data.flatten()))
#         sectime, dt0 = dttool.convert_datetime_to_sectime(x_data, dt0=dt0)
#         x_data = sectime
#
#     # check x dim:
#     if len(x_data.shape) == 1:
#         x_dim = 1
#     elif len(x_data.shape) == 2:
#         if x_data.shape[1] == num_y:
#             x_dim = 2
#         else:
#             x_dim = 1
#     else:
#         raise ValueError
#     if x_dim == 1:
#         xd = np.tile(x_data.flatten(), (1, num_y))
#     else:
#         xd = x_data
#     min_x = np.nanmin(xd.flatten())
#     max_x = np.nanmax(xd.flatten())
#     # check y dim:
#     if len(y_data.shape) == 1:
#         y_dim = 1
#     elif len(y_data.shape) == 2:
#         if y_data.shape[1] == num_y:
#             y_dim = 2
#         else:
#             y_dim = 1
#     else:
#         raise ValueError
#     if y_dim == 1:
#         yd = np.tile(x_data.flatten(), (num_x, 1))
#     else:
#         yd = y_data
#     min_y = np.nanmin(yd.flatten())
#     max_y = np.nanmax(yd.flatten())
#
#     if x_data_res is None:
#         x_data_res_ = np.median(np.diff(xd[:, 0].flatten()))
#     else:
#         x_data_res_ = x_data_res
#     if y_data_res is None:
#         y_data_res_ = np.median(np.diff(yd[0, :].flatten()))
#     else:
#         y_data_res_ = y_data_res
#
#     if x_grid_res is None:
#         xx = xd[:, 0].flatten()
#         # along_x_interp=False
#         # along_x_binning=False
#         # x_grid_res=x_data_res
#     else:
#         if x_grid_min is None:
#             x_grid_min = np.floor((min_x / x_grid_res)) * x_grid_res
#         if x_grid_max is None:
#             x_grid_max = np.ceil((max_x / x_grid_res)) * x_grid_res
#         xx = np.arange(x_grid_min, x_grid_max+x_grid_res, x_grid_res)
#     if y_grid_res is None:
#         yy = yd[0, :].flatten()
#         # along_y_interp=False
#         # along_y_binning=False
#         # y_grid_res=y_data_res
#     else:
#         if y_grid_min is None:
#             y_grid_min = np.floor((min_y / y_grid_res)) * y_grid_res
#         if y_grid_max is None:
#             y_grid_max = np.ceil((max_y / y_grid_res)) * y_grid_res
#         yy = np.arange(y_grid_min, y_grid_max+y_grid_res, y_grid_res)
#
#     grid_x, grid_y = numpy.meshgrid(xx,yy)
#     grid_z = np.empty_like(grid_x)
#     grid_z[::] = np.nan
#
#     if along_y_interp:
#         grid_x_1 = np.empty_like((num_x, grid_z.shape[1])) * np.nan
#         grid_y_1 = np.empty_like((num_x, grid_z.shape[1])) * np.nan
#         grid_z_1 = np.empty_like((num_x, grid_z.shape[1])) * np.nan
#         for i in range(num_x):
#             x1 = xd[i, :].flatten()
#             y1 = yd[i, :].flatten()
#             z1 = z_data[i, :].flatten()
#
#             inds_finite = np.where(np.isfinite(x1) & np.isfinite(y1) & np.isfinite(z1))[0]
#             if not list(inds_finite):
#                 continue
#
#             f = interp1d(y1[inds_finite], x1[inds_finite],
#                 kind='nearest', bounds_error=False, fill_value=np.nan)
#             x_i = f(grid_y[i, :].flatten())
#             grid_x_1[i, :] = x_i
#
#             f = interp1d(y1[inds_finite], y1[inds_finite],
#                 kind='nearest', bounds_error=False, fill_value=np.nan)
#             y_i = f(grid_y[i, :].flatten())
#             grid_y_1[i, :] = y_i
#
#             f = interp1d(
#                 y1[inds_finite], z1[inds_finite],
#                 kind=along_y_interp_method, bounds_error=False, fill_value=np.nan)
#             z_i = f(grid_y[i, :].flatten())
#
#             grid_z_1[i, :] = z_i


def regridding_2d_xgaps(
    x, y, z,
    xtype=None, xres=None):
    
    x1 = x.flatten()
    if xtype == 'datetime':
        # dt0 = datetime.datetime(1970, 1, 1)
        sectime, dt0 = dttool.convert_datetime_to_sectime(x1)
        x1 = sectime
    
    diff_x = numpy.diff(x1)
    if xres is None:
        xres = numpy.median(diff_x)
        
    if xtype == 'datetime' and basic.isnumeric(xres):
        xres = datetime.timedelta(seconds=xres)
        
    xx = numpy.hstack(((x.flatten() - xres / 2)[:, numpy.newaxis], (x.flatten() + xres / 2)[:, numpy.newaxis]))
    xnew = xx.flatten()
    
    if len(y.shape) == 1 or 1 in y.shape:
        ynew = y
    else:
        yy = numpy.hstack((y, y))
        ynew = yy.reshape((xnew.shape[0], y.shape[1]))
    
    zz = numpy.hstack((z, z))
    znew = zz.reshape((xnew.shape[0], z.shape[1]))
    znew[1::2, :] = numpy.nan
    znew = znew[:-1, :]
    
    return xnew, ynew, znew
         

def data_resample_2d(
        x=None, y=None, z=None, xtype=None, xres=None, xresscale=3,
        method='Null',  # Null - insert NaN, 'linear', 'cubic', ... (interpolation method)
        axis=0, forward=True, depth=0
):

    x1 = x
    if xtype == 'datetime':
        # dt0 = datetime.datetime(1970, 1, 1)
        sectime, dt0 = dttool.convert_datetime_to_sectime(x1)
        x1 = sectime

    diff_x1 = numpy.diff(x1.flatten())
    if xres is None:
        xres = numpy.median(diff_x1)
    inds = numpy.where(diff_x1 > xres * xresscale)[0]

    if len(inds) == 0:
        return x, y, z

    if depth > 10:
        return x, y, z

    inds = [i+1 for i in inds]
    # for x
    value = []
    for ind in inds:
        res = xres
        if xtype == 'datetime':
            res = datetime.timedelta(seconds=xres)
        if forward:
            value.append(x[ind - 1] + res/10)
        else:
            value.append(x[ind] - res/10)
    xnew = numpy.insert(x, inds, value, axis=axis)

    # for y
    method_y = 'nearest'
    ifunc = interp1d(x1.flatten(), y, kind=method_y, axis=axis, fill_value="extrapolate")
    x_p = xnew
    if xtype == 'datetime':
        x_p, dt0 = dttool.convert_datetime_to_sectime(xnew)
    ynew = ifunc(x_p.flatten())

    # for z
    if method == 'Null':
        value = numpy.nan
        znew = numpy.insert(z, inds, value, axis=axis)
    else:
        ifunc = interp1d(x1.flatten(), z, kind=method, axis=axis)
        x_p = xnew
        if xtype == 'datetime':
            x_p, dt0 = dttool.convert_datetime_to_sectime(xnew)
        znew = ifunc(x_p.flatten())
    if method == 'Null':
        xnew, ynew, znew = data_resample_2d(
            x=xnew, y=ynew, z=znew, xtype=xtype, xres=xres, xresscale=xresscale,
            method='Null',  # Null - insert NaN, 'linear', 'cubic', ... (interpolation method)
            axis=axis, forward=False, depth=depth+1
        )

    return xnew, ynew, znew



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import numpy as np
import datetime
import cartopy.crs as ccrs
import re
import copy
from cartopy.mpl.ticker import (
    LongitudeLocator, LatitudeLocator,
    LongitudeFormatter, LatitudeFormatter)
import matplotlib.ticker as mticker
import matplotlib.path as mpath
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.cm as mcm
from scipy.interpolate import interp1d, griddata
import matplotlib.cm as cm
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER

import geospacelab.visualization.mpl as mpl
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.numpymath as mathtool
import geospacelab.toolbox.utilities.pybasic as pybasic
from geospacelab.visualization.mpl.geomap.__base__ import GeoPanelBase
from geospacelab.datahub.__variable_base__ import VariableBase


class GeoPanel(GeoPanelBase):

    def __init__(self, *args, proj_class=None, proj_config: dict = None, **kwargs):
        super().__init__(*args, proj_class=proj_class, proj_config=proj_config, **kwargs)


class PolarMapPanel(GeoPanel):

    def __init__(self, *args, cs='AACGM', style=None, pole='N', lon_c=None, lst_c=None, mlt_c=None,
                 mlon_c=None, ut=None,
                 boundary_lat=30., boundary_style='circle',
                 mirror_south=False,
                 proj_type='Stereographic', **kwargs):
        if style is None:
            style = input("Specify the mapping style: lon-fixed, lst-fixed, mlon-fixed, or mlt-fixed? ")
        self.style = style
        if style in ['lon-fixed']:
            if lon_c is None:
                raise ValueError
            lst_c = None
            mlon_c = None
            mlt_c = None

        if style in ['lst-fixed']:
            if lst_c is None:
                raise ValueError
            lon_c = None
            mlon_c = None
            mlt_c = None

        if style in ['mlon-fixed']:
            if mlon_c is None:
                raise ValueError
            lst_c = None
            lon_c = mlon_c
            mlt_c = None

        if style in ['mlt-fixed']:
            if mlt_c is None:
                raise ValueError
            lst_c = None
            lon_c = None
            mlon_c = None

        # if lon_c is not None and pole == 'S':
        #    lon_c = lon_c + 180.

        self.pole = pole
        self.lat_c = kwargs.pop('lat_c', None)
        self.lon_c = lon_c
        self.ut = ut
        self.boundary_lat = boundary_lat
        self.boundary_style = boundary_style
        self.lst_c = lst_c
        self.cs = cs
        self.depend_mlt = False
        self.mlt_c = mlt_c
        self.mirror_south = mirror_south
        self._extent = None
        self._sector = kwargs.pop('sector', 'off')
        self.boundary_latitudes = None
        self.boundary_longitudes = None
        proj_class = getattr(ccrs, proj_type)
        proj_config = {'central_latitude': self.lat_c, 'central_longitude': self.lon_c}
        super().__init__(*args, proj_class=proj_class, proj_config=proj_config, **kwargs)

        if self._sector == 'off':
            self.set_map_extent()
            self.set_map_boundary()
            self._check_mirror_south()

    @staticmethod
    def _transform_mlt_to_lon(mlt):
        lon = mlt / 24. * 360.
        lon = np.mod(lon, 360.)
        return lon

    def _transform_lst_to_lon(self, lst):
        if self.pole == 'N':
            lon = np.mod((lst - (self.ut.hour + self.ut.minute / 60)) * 15., 360.)
        elif self.pole == 'S':
            lon = np.mod((lst - (self.ut.hour + self.ut.minute / 60)) * 15. + 180., 360.)
        return lon

    # def add_lands(self):
    #     import cartopy.io.shapereader as shpreader
    #
    #     resolution = '110m'
    #     shpfilename = shpreader.natural_earth(resolution=resolution,
    #                                           category='physical',
    #                                           name='coastline')
    #     reader = shpreader.Reader(shpfilename)
    #     lands = list(reader.geometries())
    #
    #     for ind1, land in enumerate(lands):
    #         land_polygons = list(land)
    #         for ind2, polygon in enumerate(land_polygons):
    #             x, y = polygon.exterior.coords.xy
    #
    def cs_transform(self, cs_fr=None, cs_to=None, coords=None, ut=None):
        import geospacelab.cs as geo_cs

        if cs_to is None:
            cs_to = self.cs
        if ut is None:
            ut = self.ut
        cs_class = getattr(geo_cs, cs_fr.upper())
        cs1 = cs_class(coords=coords, ut=ut)
        if cs_fr != cs_to:
            cs2 = cs1(cs_to=cs_to, append_mlt=self.depend_mlt)
        else:
            cs2 = cs1
        if self.style == 'lst-fixed':
            try:
                if cs2.coords.lst is None:
                    ts = np.array([
                        t.hour + t.minute/60 + t.second/3600 for t in ut
                    ])
                    cs2.coords.lst = (cs2.coords.lon / 15. + ts) % 24.
                cs2.coords.lon = self._transform_lst_to_lon(cs2.coords.lst)
            except:
                mylog.StreamLogger.warning("LST is not specified!")
        if self.depend_mlt:
            cs2.coords.lon = self._transform_mlt_to_lon(cs2.coords.mlt)
        return cs2

    def overlay_coastlines(self, linestyle='-', linewidth=0.5, color='#797A7D', zorder=100, alpha=0.7,
                           resolution='110m', **kwargs):
        import cartopy.io.shapereader as shpreader

        shpfilename = shpreader.natural_earth(resolution=resolution,
                                              category='physical',
                                              name='coastline')
        coastlines = list(shpreader.Reader(shpfilename).geometries())

        x = np.array([])
        y = np.array([])
        for ind, c in enumerate(coastlines[:-1]):
            # print(ind)
            # print(len(c.xy[0]))
            # if ind not in [4013, 4014]:
            #    continue
            try:
                x0 = np.array(c.xy[0])
                y0 = np.array(c.xy[1])
            except:
                continue
            # if len(x0) < 20:  # omit small islands, etc.
            #    continue
            x0 = np.mod(x0[::1], 360)
            y0 = y0[::1]
            x = np.append(np.append(x, x0), np.nan)
            y = np.append(np.append(y, y0), np.nan)

            # csObj = scs.SpaceCS(x, y, CS='GEO', dt=self.dt, coords_labels=['lon', 'lat'])

        coords = {'lat': y, 'lon': x, 'height': np.ones_like(y) * 0}
        coords = self.cs_transform(cs_fr='GEO', cs_to=self.cs, coords=coords)
        x_new = coords['lon']
        y_new = coords['lat']
        # x_new, y_new = x, y
        self().plot(
            x_new, y_new,
            transform=ccrs.Geodetic(),
            linestyle=linestyle, linewidth=linewidth, color=color, zorder=zorder, alpha=alpha,
            **kwargs)
        # self.ax.scatter(x_new, y_new, transform=self.default_transform,
        #             marker='.', edgecolors='none', color='#C0C0C0', s=1)

        return

    def overlay_gridlines(self,
                      lat_res=None, lon_res=None,
                      lat_label_separator=None, lon_label_separator=None,
                      lat_label_format=None, lon_label_format=None,
                      lat_label=True, lat_label_clock=6.5, lat_label_config={},
                      lon_label=True, lon_label_config={},
                      gridlines_config={}):

        default_gridlines_label_config = {
            'lon-fixed': {
                'lat_res': 10.,
                'lon_res': 30.,
                'lat_label_separator': 0,
                'lon_label_separator': 0,
                'lat_label_format': '%d%D%N',
                'lon_label_format': '%d%D%E',
            },
            'mlon-fixed': {
                'lat_res': 10.,
                'lon_res': 30.,
                'lat_label_separator': 0,
                'lon_label_separator': 0,
                'lat_label_format': '%d%D%N',
                'lon_label_format': '%d%D%E',
            },
            'lst-fixed': {
                'lat_res': 10.,
                'lon_res': 15.,
                'lat_label_separator': 0,
                'lon_label_separator': 2,
                'lat_label_format': '%d%D%N',
                'lon_label_format': '%02d LT',
            },
            'mlt-fixed': {
                'lat_res': 10.,
                'lon_res': 15.,
                'lat_label_separator': 0,
                'lon_label_separator': 2,
                'lat_label_format': '%d%D%N',
                'lon_label_format': '%02d MLT',
            },
        }

        def label_formatter(value, fmt=''):
            """
            This will be deprecated in V1.0. Use the python string format instead.
            """
            ms = re.findall(r"(%[.0-9]*[a-zA-Z])", fmt)
            fmt_new = copy.deepcopy(fmt)
            patterns = []
            if "%N" in ms:
                if value > 0:
                    lb_ns = 'N'
                elif value < 0:
                    lb_ns = 'S'
                else:
                    lb_ns = ''
                value = np.abs(value)

            if "%E" in ms:
                mod_value = np.mod(value - 180, 360) - 180.
                if mod_value == 0 or mod_value == 180:
                    lb_ew = ''
                elif mod_value < 0:
                    lb_ew = 'W'
                else:
                    lb_ew = 'E'
                value = np.abs(mod_value)

            for ind_m, m in enumerate(ms):
                if m == "%N":
                    fmt_new = fmt_new.replace(m, '{}')
                    patterns.append(lb_ns)
                elif m == "%E":
                    fmt_new = fmt_new.replace(m, '{}')
                    patterns.append(lb_ew)
                elif m == "%D":
                    continue
                else:
                    if 'd' in m:
                        value = int(value)
                    elif 'f' in m:
                        value = value
                    else:
                        raise NotImplementedError
                    fmt_new = fmt_new.replace(m, '{:' + m[1:] + '}')
                    patterns.append(value)
            string_new = fmt_new.format(*patterns)
            if "%D" in string_new:
                splits = string_new.split("%D")
                string_new = splits[0] + r'$^{\circ}$' + splits[1]

            return string_new

        default_label_config = default_gridlines_label_config[self.style]
        if lat_res is None:
            lat_res = default_label_config['lat_res']
        if lon_res is None:
            lon_res = default_label_config['lon_res']
        if lat_label_separator is None:
            lat_label_separator = default_label_config['lat_label_separator']
        if lon_label_separator is None:
            lon_label_separator = default_label_config['lon_label_separator']
        if lat_label_format is None:
            lat_label_format = default_label_config['lat_label_format']
        if lon_label_format is None:
            lon_label_format = default_label_config['lon_label_format']

        # xlocator
        if self.style == 'lst-fixed':
            lon_shift = (self.ut - datetime.datetime(
                self.ut.year, self.ut.month, self.ut.day, self.ut.hour
            )).total_seconds() / 86400. * 360
        else:
            lon_shift = 0
        num_lons = 360. // lon_res + 1.
        lons = np.linspace(-180., 180., int(num_lons)) - lon_shift
        xlocator = mticker.FixedLocator(lons)

        # ylocator
        lat_fr = np.abs(self.boundary_lat) + np.mod(90 - np.abs(self.boundary_lat), lat_res)
        lat_to = 90.
        # num_lats = (lat_to - lat_fr) / lat_res + 1.
        lats = np.arange(lat_fr, lat_to, lat_res) * np.sign(self.lat_c)
        ylocator = mticker.FixedLocator(lats)

        pybasic.dict_set_default(gridlines_config, color='#331900', linewidth=0.5, linestyle=':',
                                     draw_labels=False)
        gl = self().gridlines(crs=ccrs.PlateCarree(), **gridlines_config,)

        gl.xlocator = xlocator
        gl.ylocator = ylocator

        if lat_label:
            pybasic.dict_set_default(lat_label_config, fontsize=plt.rcParams['font.size'] - 2, fontweight='roman',
                    ha='center', va='center',
                    family=plt.rcParams['font.family'], alpha=0.9
                )
            if self.pole == 'S':
                clock = lat_label_clock / 12 * 360
                rotation = 180 - clock
                if self.mirror_south:
                    clock = - clock
                    rotation = 180 + clock
            else:
                clock = np.mod(180. - lat_label_clock / 12 * 360, 360)
                rotation = clock
            label_lon = self.lon_c + clock

            lat_lim = [np.min(np.abs(self.boundary_latitudes)), np.max(np.abs(self.boundary_latitudes))]
            if lat_lim[0] == lat_lim[1]:
                lat_lim[1] = 90
            for ind, lat in enumerate(lats):
                if abs(lat) < lat_lim[0]:
                    continue
                if abs(lat) > lat_lim[1]:
                    continue
                if np.mod(ind, lat_label_separator+1) != 0:
                    continue
                # if lat > 0:
                #     label = r'' + '{:d}'.format(int(lat)) + r'$^{\circ}$N'
                # elif lat < 0:
                #     label = r'' + '{:d}'.format(int(-lat)) + r'$^{\circ}$S'
                # else:
                #     label = r'' + '{:d}'.format(int(0)) + r'$^{\circ}$S'
                if '%' in lat_label_format:
                    label = label_formatter(lat, fmt=lat_label_format)
                else:
                    label = lat_label_format.format(lat)
                self().text(
                    label_lon, lat, label, transform=ccrs.PlateCarree(),
                    rotation=rotation,
                    **lat_label_config
                )

        if lon_label:
            pybasic.dict_set_default(
                lon_label_config,
                va='center', ha='center',
                family=plt.rcParams['font.family'], fontweight='ultralight')
            lb_lons = np.arange(0, 360., lon_res)
            if self.pole == 'S':
                lon_shift_south = 180.
            else:
                lon_shift_south = 0
            if self.style == 'lon-fixed' or self.style == 'mlon-fixed':
                lb_lons_loc = lb_lons
            elif self.style == 'lst-fixed':
                lb_lons_loc = np.mod(self.lon_c + (lb_lons - self.lst_c / 24 * 360) + lon_shift_south, 360)
            elif self.style == 'mlt-fixed':
                lb_lons_loc = lb_lons

            if self.boundary_style == 'circle':
                lb_lats = np.empty_like(lb_lons)
                lb_lats[:] = self.boundary_lat
                data = self.projection.transform_points(ccrs.PlateCarree(), lb_lons_loc, lb_lats)
                xdata = data[:, 0]
                ydata = data[:, 1]
                scale = (self._extent[1] - self._extent[0]) * 0.03
                xdata = xdata + scale * np.sin((lb_lons_loc - self.lon_c) * np.pi / 180.)
                ydata = ydata - np.sign(self.lat_c) * scale * np.cos((lb_lons_loc - self.lon_c) * np.pi / 180.)
                for ind, lb_lon in enumerate(lb_lons):
                    if np.mod(ind, lon_label_separator+1) != 0:
                        continue
                    x = xdata[ind]
                    y = ydata[ind]
                    # if self.style in ['lon-fixed', 'mlon-fixed']:
                    #     lb_lon = np.mod(lb_lon + 180, 360) - 180
                    #     if lb_lon == 0 or np.abs(lb_lon) == 180.:
                    #         label = r'' + '{:d}'.format(int(np.abs(lb_lon))) + r'$^{\circ}$'
                    #     if lb_lon < 0:
                    #         label = r'' + '{:d}'.format(int(-lb_lon)) + r'$^{\circ}$W'
                    #     else:
                    #         label = r'' + '{:d}'.format(int(lb_lon)) + r'$^{\circ}$E'
                    # elif self.style == 'mlt-fixed':
                    #     label = '{:d}'.format(int(lb_lon / 15)) + ' MLT'
                    # elif self.style == 'lst-fixed':
                    #     label = '{:d}'.format(int(lb_lon / 15)) + ' LT'
                    if self.style in ['lon-fixed', 'mlon-fixed']:
                        lb_value = lb_lon
                    elif self.style in ['lst-fixed', 'mlt-fixed']:
                        lb_value = lb_lon / 15
                    label = label_formatter(lb_value, fmt=lon_label_format)
                    if self.pole == 'S':
                        rotation = self.lon_c - lb_lons_loc[ind]
                        if self.mirror_south:
                            rotation = -self.lon_c + lb_lons_loc[ind]
                    else:
                        rotation = (lb_lons_loc[ind] - self.lon_c) + 180
                    self().text(
                        x, y, label,
                        rotation=rotation,
                        **lon_label_config,
                    )

        # gl.xformatter = LONGITUDE_FORMATTER()
        # gl.yformatter = LATITUDE_FORMATTER()
        # for ea in gl.label_artists:
        #     if ea[1]==False:
        #         tx = ea[2]
        #         xy = tx.get_position()
        #         #print(xy)
        #
        #         tx.set_position([30, xy[1]])

    def set_map_extent(self, boundary_latitudes=None, boundary_longitudes=None, **kwargs):
        if any([a is None for a in [boundary_latitudes, boundary_longitudes]]):
            boundary_longitudes = np.arange(0., 360., 5.)
            boundary_latitudes = np.empty_like(boundary_longitudes)
            boundary_latitudes[:] = 1. * self.boundary_lat
        self.boundary_latitudes = boundary_latitudes
        self.boundary_longitudes = boundary_longitudes
        super().set_map_extent(boundary_latitudes, boundary_longitudes)

    def set_map_boundary(self, path=None, transform=None, **kwargs):
        if any([a is None for a in [path, transform]]):
            style = self.boundary_style
            if style == 'square':
                return
            elif style == 'circle':
                theta = np.linspace(0, 2 * np.pi, 100)
                center, radius = [0.5, 0.5], 0.5
                verts = np.vstack([np.sin(theta), np.cos(theta)]).T
                path = mpath.Path(verts * radius + center)
                transform = self().transAxes
            else:
                raise NotImplementedError

        super().set_map_boundary(path, transform)

    def _check_mirror_south(self):
        if self.pole == 'S' and self.mirror_south:
            xlim = self.major_ax.get_xlim()
            self.major_ax.set_xlim([max(xlim), min(xlim)])

    def _set_boundary_style(self):
        """
        Deprecated.
        :return:
        """
        style = self.boundary_style
        if style == 'square':
            return
        elif style == 'circle':
            theta = np.linspace(0, 2 * np.pi, 400)
            center = self.projection.transform_point(self.lon_c, self.lat_c, ccrs.PlateCarree())
            radius = (self._extent[1] - self._extent[0]) / 2
            verts = np.vstack([np.sin(theta), np.cos(theta)]).T
            circle = mpath.Path(verts * radius + center)
            self().set_boundary(circle, transform=self.projection)
        else:
            raise NotImplementedError

    def overlay_pcolormesh(
            self, data=None, coords=None, cs=None, *,
            ut = None,
            regridding: bool = True,
            data_res: float = 0.7,
            grid_res: float = 0.1,
            interp_method: str = 'linear',
            c_scale='linear', c_lim=None,
            **kwargs):

        kwargs.setdefault('shading', 'auto')
        if c_lim is None:
            c_lim = [0, 0]
            c_lim[0] = np.nanmin(data.flatten())
            c_lim[1] = np.nanmax(data.flatten())

        if c_scale == 'log':
            norm = mcolors.LogNorm(vmin=c_lim[0], vmax=c_lim[1])
            kwargs.update(norm=norm)
        else:
            kwargs.update(vmin=c_lim[0])
            kwargs.update(vmax=c_lim[1])

        cs_new = self.cs_transform(cs_fr=cs, coords=coords, ut=ut)
        lat_in = cs_new['lat']
        lon_in = cs_new['lon']
        data_in = data
        if regridding:
            lon_in, lat_in, data_in = self.grid_data(lon_in, lat_in, data_in, data_res=data_res, grid_res=grid_res)
            transform = None
            # self.boundary_lat
            #
            # grid_lat_res = kwargs.pop('grid_lat_res', 0.1)
            # grid_lon_res = kwargs.pop('grid_lon_res', 0.1)
            #
            # ind_data = np.where(np.isfinite(data))
            # data_pts = data[ind_data]
            # lat_pts = cs_new['lat'][ind_data]
            # lon_pts = cs_new['lon'][ind_data]
            #
            # grid_lon, grid_lat = np.meshgrid(
            #     np.arange(0., 360., grid_lon_res),
            #     np.append(np.arange(self.boundary_lat, self.lat_c, np.sign(self.lat_c) * grid_lat_res), self.lat_c)
            # )
            #
            # grid_data = griddata((lon_pts, lat_pts), data_pts, (grid_lon, grid_lat), method='nearest')
            # grid_data_lat = griddata(
            #     (lon_pts, lat_pts), lat_pts, (grid_lon, grid_lat), method='nearest'
            # )
            # grid_data_lon = griddata(
            #     (lon_pts, lat_pts), lon_pts, (grid_lon, grid_lat), method='nearest'
            # )
            # factor = np.pi / 180.
            # big_circle_d = 6371. * np.arccos(
            #     np.sin(grid_data_lat * factor) * np.sin(grid_lat * factor) +
            #     np.cos(grid_data_lat * factor) * np.cos(grid_lat * factor) * np.cos(
            #         np.abs((grid_lon - grid_data_lon) * factor))
            # )
            # grid_data = np.where(big_circle_d < data_lat_res*111.2, grid_data, np.nan)
            # lat_in = grid_lat
            # lon_in = grid_lon
            # data_in = grid_data
        else:
            transform = ccrs.PlateCarree()

        ipm = self().pcolormesh(lon_in, lat_in, data_in, transform=transform, **kwargs)

        return ipm

    def overlay_contour(
            self,
            data, coords=None, cs=None,
            regridding=True, grid_res=0.1, interp_method='linear', sparsely=False, data_res=0.5,
            **kwargs,
    ):
        levels = kwargs.pop('levels', 10)
        if cs is None:
            cs = self.cs

        cs_new = self.cs_transform(cs_fr=cs, coords=coords)
        lat_in = cs_new['lat']
        lon_in = cs_new['lon']
        data_in = data
        if regridding:
            lon_in, lat_in, data_in = self.grid_data(
                lon_in, lat_in, data_in, sparsely=sparsely,
                grid_res=grid_res, data_res=data_res, interp_method=interp_method)
            transform = None
        else:
            transform = ccrs.PlateCarree()

        ict = self.major_ax.contour(lon_in, lat_in, data_in, levels, transform=transform, **kwargs)
        return ict

    def grid_data(self, lon_in, lat_in, data_in, grid_res=0.1, sparsely=True, data_res=0.5, interp_method='linear'):
        if self.pole == "N":
            sign_lat = 1.
        else:
            sign_lat = -1.
        ind_data = np.where(np.isfinite(data_in) & (lat_in * sign_lat > (self.boundary_lat-5.)) & np.isfinite(lon_in))
        data_pts = data_in[ind_data]
        lat_pts = lat_in[ind_data]
        lon_pts = lon_in[ind_data]

        pos_data = self.projection.transform_points(ccrs.PlateCarree(), lon_pts, lat_pts)
        pos_x = np.float32(pos_data[:, 0])
        pos_y = np.float32(pos_data[:, 1])

        lim=[]
        xlim = list(self.major_ax.get_xlim())
        ylim = list(self.major_ax.get_ylim())
        lim.extend(xlim)
        lim.extend(ylim)
        lim = [np.float32(x) for x in lim]
        width = np.max(lim) - np.min(lim)
        lat_width = (90. - np.abs(self.boundary_lat)) * 2
        grid_x_res = grid_res / lat_width * width
        if xlim[0] > xlim[1]:
            grid_x_res = -grid_x_res
        grid_x, grid_y = np.meshgrid(
            np.arange(lim[0], lim[1], grid_x_res),
            np.arange(lim[0], lim[1], grid_x_res)
        )

        grid_data = griddata((pos_x, pos_y), data_pts, (grid_x, grid_y), method=interp_method)
        grid_data_x = griddata(
            (pos_x, pos_y), pos_x, (grid_x, grid_y), method='nearest'
        )
        grid_data_y = griddata(
            (pos_x, pos_y), pos_y, (grid_x, grid_y), method='nearest'
        )

        if sparsely:
            ind_out = np.where(
                np.sqrt((grid_data_x - grid_x) ** 2 + (grid_data_y - grid_y) ** 2) > data_res / lat_width * width
            )
            grid_data[ind_out] = np.nan

        return grid_x, grid_y, grid_data

    def overlay_line(self, ut=None, coords=None, cs=None, *, color='#EEEEEE', linewidth=1., linestyle='--', **kwargs):
        transform = kwargs.pop('transform', ccrs.Geodetic())
        cs_new = self.cs_transform(cs_fr=cs, coords=coords, ut=ut)

        if self.pole == 'N':
            ind_lat = np.where(cs_new['lat'] > self.boundary_lat)[0]
        else:
            ind_lat = np.where(cs_new['lat'] < self.boundary_lat)[0]

        lat_in = cs_new['lat'][ind_lat]
        lon_in = cs_new['lon'][ind_lat]

        self.major_ax.plot(lon_in, lat_in, transform=ccrs.Geodetic(), color=color, linewidth=linewidth, linestyle=linestyle, **kwargs)

    def overlay_vectors(
            self, vectors, coords=None, cs=None, *, ut=None, unit_vector=None,
            unit_vector_scale=0.1, vector_unit='',
            color='r', alpha=0.9, quiverkey_config={},
            vector_width=1,
            vector_root_marker='o',
            vector_root_marker_size=0.2,
            vector_root_marker_color='k',
            vector_root_marker_facecolor='k',
            shading='on',
            edge='off',
            edge_color=None,
            edge_linestyle='-',
            edge_linewidth=1.,
            edge_marker=None,
            edge_markersize=5,
            edge_alpha=0.8,
            legend='on',
            legend_pos_x=0.9,
            legend_pos_y=0.95,
            legend_label=None,
            legend_linewidth=None,
            legend_fontsize=10,
            legend_color=None,
            **kwargs):
        kwargs.setdefault('zorder', 50)
        if ut is None:
            ut = self.ut

        cmap = kwargs.pop('cmap', None)

        cs_new = self.cs_transform(cs_fr=cs, coords=coords, ut=ut)

        if self.pole == 'N':
            ind_lat = np.where(cs_new['lat'] > self.boundary_lat)[0]
        else:
            ind_lat = np.where(cs_new['lat'] < self.boundary_lat)[0]

        lat_in = cs_new['lat'][ind_lat]
        lon_in = cs_new['lon'][ind_lat]
        data = self.projection.transform_points(ccrs.PlateCarree(), lon_in, lat_in)
        xdata = np.float32(data[:, 0])
        ydata = np.float32(data[:, 1])
        # sign = np.sign(xdata[1] - xdata[0])

        v_N_in = vectors[0][ind_lat]
        v_E_in = vectors[1][ind_lat]
        # v_mag_in = np.sqrt(v_N_in ** 2 + v_E_in ** 2)

        width = np.float32(self._extent[1] - self._extent[0])
        data_scale = 1 / unit_vector * width * unit_vector_scale
        v_N_in = v_N_in * data_scale
        v_E_in = v_E_in * data_scale


        pseudo_lons = (lon_in - self.lon_c) * np.pi / 180.
        xq = xdata
        yq = ydata
        uq = - v_N_in * np.sin(pseudo_lons) + v_E_in * np.cos(pseudo_lons)
        vq = v_N_in * np.cos(pseudo_lons) + v_E_in * np.sin(pseudo_lons)

        kwargs.update(visible=True)
        if vector_root_marker is not None:
            self.major_ax.plot(
                xq, yq,
                linestyle='',
                marker=vector_root_marker,
                markersize=vector_root_marker_size,
                color=vector_root_marker_color,
                markerfacecolor=vector_root_marker_facecolor, zorder=kwargs['zorder'])

        mag = np.hypot(uq, vq)

        if cmap is not None:
            if isinstance(cmap, str):
                cmap = mcm.get_cmap(cmap)
            clim = kwargs.pop('clim', None)

            if clim is None:
                norm = mcolors.Normalize()
                norm.autoscale(mag)
            else:
                clim = np.array(clim) / unit_vector * width * unit_vector_scale
                norm = mcolors.Normalize(vmin=clim[0], vmax=clim[1])
            colors = cmap(norm(mag))
            args = (xq, yq, uq, vq)
        else:
            colors = color
            args = (xq, yq, uq, vq)

        iq = self.major_ax.quiver(
            *args,
            units='xy', angles='xy', scale=1., scale_units='xy',
            width=vector_width * 0.002 * (self._extent[1] - self._extent[0]),
            headlength=0, headaxislength=0, pivot='tail', color=colors, alpha=alpha, cmap=cmap, **kwargs
        )
        iq.data_scale = data_scale
        # Add quiverkey
        if legend in ['on', True]:
            if legend_linewidth is None:
                legend_linewidth = vector_width
            if legend_label is None:
                legend_label = str(unit_vector) + ' ' + vector_unit
            quiverkey_config = pybasic.dict_set_default(
                quiverkey_config, X=legend_pos_x, Y=legend_pos_y, U=width*unit_vector_scale,
                width=legend_linewidth * 0.002 * (self._extent[1] - self._extent[0]),
                label=legend_label, color=color,
            )
            X = quiverkey_config.pop('X')
            Y = quiverkey_config.pop('Y')
            U = quiverkey_config.pop('U')
            label = quiverkey_config.pop('label')
            fontproperties = kwargs.pop('fontproperties', {})
            fontproperties.update(size=legend_fontsize)
            kwargs.update(visible='on')
            iqk = self.major_ax.quiverkey(
                iq, X, Y, U, label, color=legend_color, fontproperties=fontproperties, **kwargs
            )

        return iq

    def overlay_sc_trajectory(self, sc_ut=None, sc_coords=None, cs=None, *, show_trajectory=True, color='#EEEEEE',
                          time_tick=True, time_tick_res=300., time_tick_scale=0.05,
                          time_tick_label=True, time_tick_label_format="%H:%M", time_tick_label_fontsize=8,
                          time_tick_label_rotation=45., time_tick_label_offset=0.05, 
                          time_tick_label_fontweight='normal',
                          time_major_ticks=None,
                          time_minor_tick=True, time_minor_tick_res=60,
                          time_tick_width=1, **kwargs):
        default_trajectory_config = {
            'linewidth': 1,
            'linestyle': '-',
            'color': 'k',
            'zorder': max([_.zorder for _ in self.major_ax.get_children()]),
            'alpha': 1,
        }
        default_trajectory_config.update(**kwargs.setdefault('trajectory_config', {}))
        default_trajectory_config.update(color=color)
        kwargs['trajectory_config'] = default_trajectory_config

        #
        # kwargs.setdefault('trajectory_config', {
        #     'linewidth': 1,
        #     'linestyle': '-',
        #     'color': 'k',
        #     'zorder': max([_.zorder for _ in self.major_ax.get_children()])
        # })
        # kwargs['trajectory_config'].update(color=color)

        cs_new = self.cs_transform(cs_fr=cs, coords=sc_coords, ut=sc_ut)

        if self.pole == 'N':
            ind_lat = np.where(cs_new['lat'] > self.boundary_lat)[0]
        else:
            ind_lat = np.where(cs_new['lat'] < self.boundary_lat)[0]
        
        inds_gaps = np.where(np.diff(ind_lat) > 1)[0]

        if not list(inds_gaps):
            ind_lat_segments = [[ind_lat[0], ind_lat[-1]]]
        else:
            ngaps = len(inds_gaps)

            ind_lat_segments = [[ind_lat[0], ind_lat[inds_gaps[0]]]]
            seg_body = [[ind_lat[inds_gaps[i - 1] + 1], ind_lat[inds_gaps[i]]] for i in np.arange(1, ngaps)]
            ind_lat_segments.extend(seg_body)
            ind_lat_segments.extend([[ind_lat[inds_gaps[-1] + 1], ind_lat[-1]]])

        for i_st, i_ed in ind_lat_segments:
            lat_in = cs_new['lat'][i_st:i_ed+1]
            lon_in = cs_new['lon'][i_st:i_ed+1]
            dts_in = sc_ut[i_st:i_ed+1]

            if self.ut is None:
                self.ut = sc_ut[0]
            if show_trajectory:
                self.major_ax.plot(lon_in, lat_in, transform=ccrs.Geodetic(), **kwargs['trajectory_config'])

            if time_tick:
                data = self.projection.transform_points(ccrs.PlateCarree(), lon_in, lat_in)
                xdata = data[:, 0]
                ydata = data[:, 1]
                sectime, dt0 = dttool.convert_datetime_to_sectime(
                    dts_in, datetime.datetime(self.ut.year, self.ut.month, self.ut.day)
                )
                if type(time_major_ticks) in [list, np.ndarray]:
                    time_ticks, dt0 = dttool.convert_datetime_to_sectime(
                        np.array(time_major_ticks), datetime.datetime(self.ut.year, self.ut.month, self.ut.day)
                    )
                else:
                    time_ticks = np.arange(np.ceil(sectime[0] / time_tick_res) * time_tick_res,
                                        np.ceil(sectime[-1] / time_tick_res) * time_tick_res, time_tick_res)

                f = interp1d(sectime, xdata, fill_value='extrapolate')
                x_i = f(time_ticks)
                f = interp1d(sectime, ydata, fill_value='extrapolate')
                y_i = f(time_ticks)

                slope = mathtool.calc_curve_tangent_slope(xdata, ydata)
                f = interp1d(sectime, slope, fill_value='extrapolate')
                slope_i = f(time_ticks)
                l = np.empty_like(x_i)
                l[:] = (self._extent[1] - self._extent[0]) * time_tick_scale

                xq = x_i
                yq = y_i
                uq1 = - l * np.sin(slope_i)
                vq1 = l * np.cos(slope_i)

                zorder = kwargs['trajectory_config']['zorder']
                alpha = kwargs['trajectory_config']['alpha']
                self.major_ax.quiver(
                    xq, yq, uq1, vq1,
                    units='xy', angles='xy', scale=1., scale_units='xy',
                    width=time_tick_width*0.003 * (self._extent[1] - self._extent[0]),
                    headlength=0, headaxislength=0, pivot='middle', color=color, alpha=alpha,
                    zorder=zorder
                )

                if time_tick_label:
                    offset = time_tick_label_offset * (self._extent[1] - self._extent[0])
                    for ind, time_tick in enumerate(time_ticks):
                        time = dt0 + datetime.timedelta(seconds=int(time_tick))
                        x_time_tick = x_i[ind] - offset * np.sin(slope_i[ind])
                        y_time_tick = y_i[ind] + offset * np.cos(slope_i[ind])

                        self.major_ax.text(
                            x_time_tick, y_time_tick, time.strftime(time_tick_label_format),
                            fontsize=time_tick_label_fontsize, fontweight=time_tick_label_fontweight,
                            rotation=slope_i[ind] * 180. / np.pi + time_tick_label_rotation,
                            ha='center', va='center', color=color, alpha=alpha,
                            zorder=zorder
                        )

                # self.major_ax.plot(x_time_ticks, y_time_ticks, **kwargs['time_tick_config'])

                if time_minor_tick:

                    time_ticks = np.arange(np.ceil(sectime[0] / time_minor_tick_res) * time_minor_tick_res,
                                        np.ceil(sectime[-1] / time_minor_tick_res) * time_minor_tick_res, time_minor_tick_res)

                    f = interp1d(sectime, xdata, fill_value='extrapolate')
                    x_i = f(time_ticks)
                    f = interp1d(sectime, ydata, fill_value='extrapolate')
                    y_i = f(time_ticks)

                    f = interp1d(sectime, slope, fill_value='extrapolate')
                    slope_i = f(time_ticks)
                    l = np.empty_like(x_i)
                    l[:] = (self._extent[1] - self._extent[0]) * time_tick_scale / 2.5

                    xq = x_i
                    yq = y_i
                    uq1 = - l * np.sin(slope_i)
                    vq1 = l * np.cos(slope_i)

                    self.major_ax.quiver(
                        xq, yq, uq1, vq1,
                        units='xy', angles='xy', scale=1., scale_units='xy',
                        width=time_tick_width*0.002*(self._extent[1] - self._extent[0]),
                        headlength=0, headaxislength=0, pivot='middle', color=color,
                        zorder=zorder
                    )

    def overlay_cross_track_vector(
            self, vector, unit_vector, sc_ut=None, sc_coords=None, cs=None, *,
            unit_vector_scale=0.1, vector_unit='',
            color='r', alpha=0.8, quiverkey_config={},
            vector_width=0.5,
            shading='on',
            edge='off',
            edge_color=None,
            edge_linestyle='-',
            edge_linewidth=1.,
            edge_marker=None,
            edge_markersize=5,
            edge_alpha=0.8,
            legend='on',
            legend_pos_x=0.9,
            legend_pos_y=0.95,
            legend_label=None,
            legend_linewidth=None,
            legend_fontsize=10,
            **kwargs):

        if edge_color is None:
            edge_color = color

        kwargs.setdefault(
            'zorder', max([_.zorder for _ in self.major_ax.get_children()])
        )

        cs_new = self.cs_transform(cs_fr=cs, coords=sc_coords, ut=sc_ut)

        if self.pole == 'N':
            ind_lat = np.where(cs_new['lat'] > self.boundary_lat)[0]
        else:
            ind_lat = np.where(cs_new['lat'] < self.boundary_lat)[0]

        lat_in = cs_new['lat'][ind_lat]
        lon_in = cs_new['lon'][ind_lat]
        dts_in = sc_ut[ind_lat]
        vector = vector[ind_lat]

        data = self.projection.transform_points(ccrs.PlateCarree(), lon_in, lat_in)
        xdata = np.float32(data[:, 0])
        ydata = np.float32(data[:, 1])
        slope = mathtool.calc_curve_tangent_slope(xdata, ydata, degree=3)

        width = np.float32(self._extent[1] - self._extent[0])
        vector = vector / unit_vector * width * unit_vector_scale
        sign = np.sign(xdata[1] - xdata[0])
        xq = xdata
        yq = ydata
        uq1 = - sign * vector * np.sin(slope)
        vq1 = sign * vector * np.cos(slope)

        if shading == 'off':
            kwargs.update(visible=True)
        else:
            kwargs.update(visible=False)
        iq = self.major_ax.quiver(
            xq, yq, uq1, vq1,
            units='xy', angles='xy', scale=1., scale_units='xy',
            width=vector_width*0.002 * (self._extent[1] - self._extent[0]),
            headlength=0, headaxislength=0, pivot='tail', color=color, alpha=alpha, **kwargs
        )
        if shading == 'on':
            xx = iq.X + iq.U
            yy = iq.Y + iq.V
            xx = np.concatenate((xx, np.flipud(xq)))
            yy = np.concatenate((yy, np.flipud(yq)))
            ishading = self.major_ax.fill(xx, yy, edgecolor=None, facecolor=color, alpha=alpha)

        # Add quiverkey
        if legend in ['on', True]:
            if legend_linewidth is None:
                legend_linewidth = vector_width
            if legend_label is None:
                legend_label = str(unit_vector) + ' ' + vector_unit
            quiverkey_config = pybasic.dict_set_default(
                quiverkey_config, X=legend_pos_x, Y=legend_pos_y, U=width*unit_vector_scale,
                width=legend_linewidth * 0.002 * (self._extent[1] - self._extent[0]),
                label=legend_label, color=color,
            )
            X = quiverkey_config.pop('X')
            Y = quiverkey_config.pop('Y')
            U = quiverkey_config.pop('U')
            label = quiverkey_config.pop('label')
            fontproperties = kwargs.pop('fontproperties', {})
            fontproperties.update(size=legend_fontsize)
            kwargs.update(visible='on')
            iqk = self.major_ax.quiverkey(
                iq, X, Y, U, label, fontproperties=fontproperties, **kwargs
            )

        if edge == 'on':
            xx = iq.X + iq.U
            yy = iq.Y + iq.V
            self.major_ax.plot(xx, yy, linestyle=edge_linestyle, linewidth=edge_linewidth,
                               color=edge_color, marker=edge_marker, markersize=edge_markersize,
                               alpha=edge_alpha)

        return iq

    def overlay_sc_coloured_line(
            self, z_data, sc_coords=None, sc_ut=None, cs=None, *, c_map='jet',
            c_scale='linear', c_lim=None, line_width=6., **kwargs):
        from matplotlib.collections import LineCollection
        import matplotlib.colors as mpl_color

        kwargs.setdefault(
            'zorder', max([_.zorder for _ in self.major_ax.get_children()])
        )
        if c_lim is None:
            c_lim = [0, 0]
            c_lim[0] = np.nanmin(z_data.flatten())
            c_lim[1] = np.nanmax(z_data.flatten())

        if c_scale == 'log':
            norm = mcolors.LogNorm(vmin=c_lim[0], vmax=c_lim[1])
        else:
            norm = mcolors.Normalize(vmin=c_lim[0], vmax=c_lim[1])

        cs_new = self.cs_transform(cs_fr=cs, coords=sc_coords, ut=sc_ut)

        if self.pole == 'N':
            ind_lat = np.where(cs_new['lat'] > self.boundary_lat)[0]
        else:
            ind_lat = np.where(cs_new['lat'] < self.boundary_lat)[0]

        lat_in = cs_new['lat'][ind_lat]
        lon_in = cs_new['lon'][ind_lat]
        dts_in = sc_ut[ind_lat]

        pos_data = self.projection.transform_points(ccrs.PlateCarree(), lon_in, lat_in)
        x = np.float32(pos_data[:, 0])
        y = np.float32(pos_data[:, 1])
        z = z_data[ind_lat]

        points = np.array([x, y]).T.reshape(-1, 1, 2)
        segments = np.concatenate([points[:-1], points[1:]], axis=1)

        lc = LineCollection(segments, norm=norm, cmap=c_map, **kwargs)
        lc.set_array(z)
        lc.set_linewidth(line_width)
        lc.set_alpha(1)
        self().add_collection(lc)
        return lc
        # clabel = label + ' (' + unit + ')'
        # self.add_colorbar(self.major_ax, line, cscale=scale, clabel=clabel)
        # cbar = plt.gcf().colorbar(line, ax=panel.major_ax, pad=0.1, fraction=0.03)
        # cbar.set_label(r'$n_e$' + '\n' + r'(cm$^{-3}$)', rotation=270, labelpad=25)

    def overlay_sites(self, site_ids=None, coords=None, cs=None, **kwargs):
        kwargs = pybasic.dict_set_default(kwargs, color='k', linestyle='', markersize=5, marker='.')
        show_site_names = kwargs.pop('show_site_names', False)
        site_names = kwargs.pop('site_names', [])
        site_name_config = kwargs.pop('site_name_config', {'fontsize': 'small', })
        site_name_config.update(color=kwargs['color'])

        cs_new = self.cs_transform(cs_fr=cs, coords=coords)
        
        isc = self().plot(cs_new['lon'], cs_new['lat'], transform=ccrs.PlateCarree(), **kwargs)
        
        if show_site_names:
            if not list(site_names):
                site_names = site_ids
            for i, s in enumerate(site_names):
                self().text(cs_new['lon'][i], cs_new['lat'][i], s, transform=ccrs.PlateCarree(), **site_name_config)
        return isc

    def add_colorbar(self, im, ax=None, figure=None,
                     c_scale='linear', c_label=None,
                     c_ticks=None, c_tick_labels=None, c_tick_label_step=1,
                     left=1.1, bottom=0.1, width=0.05, height=0.7, orientation='vertical', **kwargs
                     ):
        if figure is None:
            figure = plt.gcf()
        if ax is None:
            ax = self()
        pos = ax.get_position()
        ax_width = pos.x1 - pos.x0
        ax_height = pos.y1 - pos.y0
        ca_left = pos.x0 + ax_width * left
        ca_bottom = pos.y0 + ax_height * bottom
        ca_width = ax_width * width
        ca_height = ax_height * height
        cax = self.add_axes([ca_left, ca_bottom, ca_width, ca_height], major=False, label=c_label)
        cb = figure.colorbar(im, cax=cax, orientation=orientation, **kwargs)
        
        if orientation == 'vertical':
            lim = cax.get_ylim()
            cb.set_label(c_label, rotation=270, va='bottom', size='medium')
            axis = cb.ax.yaxis
        else:
            lim = cax.get_xlim()
            cb.set_label(c_label, rotation=0, va='top', size='medium',)
            axis = cb.ax.xaxis

        if c_ticks is not None:
            axis.set_ticks(c_ticks)
            if c_tick_labels is not None:
                axis.set_ticklabels(c_tick_labels)
        else:
            if c_scale == 'log':
                num_major_ticks = int(np.ceil(np.diff(np.log10(lim)))) * 2
                axis.set_major_locator(mticker.LogLocator(base=10.0, numticks=num_major_ticks))
                n = c_tick_label_step
                [l.set_visible(False) for (i, l) in enumerate(axis.get_ticklabels()) if i % n != 0]
                # [l.set_ha('right') for (i, l) in enumerate(cax.yaxis.get_ticklabels()) if i % n != 0]
                minorlocator = mticker.LogLocator(base=10.0, subs=(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9),
                                                     numticks=12)
                axis.set_minor_locator(minorlocator)
                axis.set_minor_formatter(mticker.NullFormatter())
        axis.set_tick_params(labelsize='small')
        return [cax, cb]

    def add_colorbar_vectors(self, iq, ax=None, figure=None,
                     c_scale='linear', c_label=None,
                     c_ticks=None, c_tick_labels=None, c_tick_label_step=1,
                     left=1.1, bottom=0.1, width=0.05, height=0.7, **kwargs
                     ):
        if figure is None:
            figure = plt.gcf()
        if ax is None:
            ax = self()
        pos = ax.get_position()
        ax_width = pos.x1 - pos.x0
        ax_height = pos.y1 - pos.y0
        ca_left = pos.x0 + ax_width * left
        ca_bottom = pos.y0 + ax_height * bottom
        ca_width = ax_width * width
        ca_height = ax_height * height
        cax = self.add_axes([ca_left, ca_bottom, ca_width, ca_height], major=False, label=c_label)
        cb = figure.colorbar(iq, cax=cax, **kwargs)
        ylim = cax.get_ylim()

        cb.set_label(c_label, rotation=270, va='bottom', size='medium')
        try:
            data_scale = iq.data_scale
        except:
            data_scale = 1
        if c_ticks is not None:
            cb.ax.yaxis.set_ticks(c_ticks)

        if c_tick_labels is not None:
            cb.ax.yaxis.set_ticklabels(c_tick_labels)
        else:
            c_ticks = cb.ax.yaxis.get_ticklocs()
            c_tick_labels = [str(tick / data_scale) for tick in c_ticks]
            cb.ax.yaxis.set_ticklabels(c_tick_labels)

        cax.yaxis.set_tick_params(labelsize='x-small')
        return [cax, cb]

    @property
    def pole(self):
        return self._pole

    @pole.setter
    def pole(self, value):
        if value in ['N', 'NORTH', 'NORTHERN']:
            self._pole = 'N'
        elif value in ['S', 'SOUTH', 'SOUTHERN']:
            self._pole = 'S'
        else:
            raise ValueError


    @property
    def lat_c(self):
        return self._lat_c

    @lat_c.setter
    def lat_c(self, value):
        if value is None:
            if self.pole == 'N':
                self._lat_c = 90.
            elif self.pole == 'S':
                self._lat_c = -90.
            else:
                raise ValueError
        else:
            self._lat_c = value

    @property
    def boundary_lat(self):
        return self._boundary_lat

    @boundary_lat.setter
    def boundary_lat(self, value):
        if self.pole == 'N':
            self._boundary_lat = np.abs(value)
        elif self.pole == 'S':
            self._boundary_lat = - np.abs(value)
        else:
            raise ValueError

    @property
    def lst_c(self):
        return self._lst_c

    @lst_c.setter
    def lst_c(self, lst):
        self._lst_c = lst
        if lst is None:
            return
        if self.pole == 'N':
            self.lon_c = np.mod((self.lst_c - (self.ut.hour + self.ut.minute / 60)) * 15., 360.)
        elif self.pole == 'S':
            self.lon_c = np.mod((self.lst_c - (self.ut.hour + self.ut.minute / 60)) * 15. + 180., 360.)


    @property
    def mlt_c(self):
        return self._mlt_c

    @mlt_c.setter
    def mlt_c(self, mlt):
        if mlt is not None:
            self.depend_mlt = True
            self.lon_c = self._transform_mlt_to_lon(mlt)
            if self.pole == 'S':
                self.lon_c = np.mod(self.lon_c + 180., 360)
            if self.cs == "GEO":
                raise AttributeError('A magnetic coordinate system must be specified (Set the attribute "cs")!')
        self._mlt_c = mlt


class PolarSectorPanel(PolarMapPanel):
    def __init__(self, *args, cs='GEO', style=None, pole='N', lat_c=None, lon_c=None, lst_c=None, mlt_c=None,
                 mlon_c=None, ut=None,
                 boundary_meridional_lim=None,
                 boundary_zonal_lim=None,
                 mirror_south=False,
                 proj_type='Stereographic', **kwargs):

        super().__init__(*args, cs=cs, style=style, pole=pole, lat_c=lat_c, lon_c=lon_c, lst_c=lst_c, mlt_c=mlt_c,
                         mlon_c=mlon_c, ut=ut,
                         sector='on',
                         mirror_south=mirror_south,
                         proj_type=proj_type, **kwargs)

        blats, blons = self._construct_boundary(boundary_zonal_lim, boundary_meridional_lim)
        self.set_map_extent(blats, blons)
        self.set_map_boundary()
        self._check_mirror_south()

    def _construct_boundary(self, boundary_zonal_lim, boundary_meridional_lim):
        import geospacelab.toolbox.utilities.numpymath as mymath

        if boundary_zonal_lim is None:
            boundary_zonal_lim = np.mod([self.lon_c - 15., self.lon_c + 15.], 360)
        if boundary_meridional_lim is None:
            boundary_meridional_lim = [self.lat_c-15, self.lat_c + 15]

        blats = np.empty((0, 1))
        blons = np.empty((0, 1))

        x = np.array([1, 2])
        y = np.array(boundary_zonal_lim)
        xq = np.linspace(1, 2, 100, endpoint=False)
        yq = mymath.interp_period_data(x, y, xq, period=360.)
        blats = np.vstack((blats, np.ones_like(yq)[:, np.newaxis] * boundary_meridional_lim[1]))
        blons = np.vstack((blons, yq[:, np.newaxis]))

        yq = np.linspace(boundary_meridional_lim[1], boundary_meridional_lim[0], 100, endpoint=False)
        blats = np.vstack((blats, yq[:, np.newaxis]))
        blons = np.vstack((blons, np.ones_like(yq)[:, np.newaxis] * boundary_zonal_lim[1]))

        x = np.array([1, 2])
        y = np.array([boundary_zonal_lim[0], boundary_zonal_lim[1]])
        xq = np.linspace(1, 2, 500, endpoint=False)
        yq = mymath.interp_period_data(x, y, xq, period=360.)
        blats = np.vstack((blats, np.ones_like(yq)[:, np.newaxis] * boundary_meridional_lim[0]))
        blons = np.vstack((blons, np.flipud(yq[:, np.newaxis])))

        yq = np.linspace(boundary_meridional_lim[0], boundary_meridional_lim[1], 500, endpoint=True)
        blats = np.vstack((blats, yq[:, np.newaxis]))
        blons = np.vstack((blons, np.ones_like(yq)[:, np.newaxis] * boundary_zonal_lim[0]))

        return blats.flatten(), blons.flatten()

    def set_map_boundary(self, path=None, transform=None, **kwargs):

        blats = self.boundary_latitudes
        blons = self.boundary_longitudes

        data = self.projection.transform_points(ccrs.PlateCarree(), blons, blats)
        verts = np.hstack((data[:, 0][:, np.newaxis], data[:, 1][:, np.newaxis]))
        path = mpath.Path(verts)

        super().set_map_boundary(path=path, transform=self.projection)





# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import datetime
import pathlib
import re

import numpy as np

import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool


class Loader(object):
    def __init__(self, file_path, direct_load=True, **kwargs):

        self.file_path = pathlib.Path(file_path)
        self.variables = {}
        self.metadata = {}

        if direct_load:
            self.load_data()

    def load_data(self):

        with open(self.file_path, 'r') as f:
            text = f.read()
            results = re.findall(
                r'^(\d{4})(\d{3})(\d{2})(\d{2})([\d.]+)\s(.{10})' +
                r'\s*(\d+)\s*(\w)\s*([\-\d.]+)\s*([\-\d.]+)\s*([\d.]+)\s*(\w)\s*(\w)' +
                r'\s*([\-\d]+)\s*([\-\d]+)\s*([\-\d]+)\s*(\d+)\.(\d+)' +
                r'\s*([\-\d]+)\s*([\-\d]+)\s*([\-\d]+)\s*(\d+)\.(\d+)\s*(\w)\n',
                text,
                re.M
            )
            results = list(zip(*results))

        seconds = [int(np.floor(float(s))) for s in results[4]]
        microseconds = [int((float(s) - np.floor(float(s)))*1e6) for s in results[4]]
        doys = np.array([int(d) for d in results[1]])
        year = int(results[0][0])
        dates = dttool.convert_doy_to_datetime(year, doys)
        hours = [int(np.floor(float(h))) for h in results[2]]
        minutes = [int(np.floor(float(m))) for m in results[3]] 
        dts = np.array([
            d + datetime.timedelta(hours=hh, minutes=mm, seconds=ss, microseconds=ms)
            for d, hh, mm, ss, ms in zip(dates, hours, minutes, seconds, microseconds)
        ])

        num_rec = len(dts)
        self.variables['SC_DATETIME'] = np.array(dts).reshape(num_rec, 1)
        self.variables['SC_GEO_ALT'] = np.array(results[10]).astype(np.float32).reshape(num_rec, 1)  # in km
        self.variables['SC_GEO_LON'] = np.array(results[9]).astype(np.float32).reshape(num_rec, 1)
        self.variables['SC_GEO_LAT'] = np.array(results[8]).astype(np.float32).reshape(num_rec, 1)
        self.variables['B_D'] = np.array(results[13]).astype(np.float32).reshape(num_rec, 1) * 1e-9
        self.variables['B_F'] = np.array(results[14]).astype(np.float32).reshape(num_rec, 1) * 1e-9 
        self.variables['B_P'] = np.array(results[15]).astype(np.float32).reshape(num_rec, 1) * 1e-9
        self.variables['d_B_D'] = np.array(results[18]).astype(np.float32).reshape(num_rec, 1) * 1e-9
        self.variables['d_B_F'] = np.array(results[19]).astype(np.float32).reshape(num_rec, 1) * 1e-9
        self.variables['d_B_P'] = np.array(results[20]).astype(np.float32).reshape(num_rec, 1) * 1e-9
        return


if __name__ == "__main__":
    load_obj = Loader(
        file_path="/home/lei/afys-data/NCEI/DMSP/SSM_MFR/F18/2024/05/PS.CKGWC_SC.U_DI.A_GP.SSMXX-F18-R99990-B9999090-APSM_AR.GLOBAL_DD.20240510_TP.000001-235958_DF.MFR"
    )


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import datetime
import numpy as np
import h5py
import re

from geospacelab.datahub.sources.madrigal.isr.pfisr.vi.loader import Loader as default_Loader
from geospacelab.datahub.sources.madrigal.isr.pfisr.vi.downloader import Downloader as default_Downloader
import geospacelab.datahub.sources.madrigal.isr.pfisr.vi.variable_config as var_config
from geospacelab.datahub.sources.madrigal import madrigal_database
from geospacelab.config import prf
from geospacelab import datahub
from geospacelab.datahub import SiteModel, DatabaseModel, FacilityModel
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog

default_dataset_attrs = {
    'kind': 'sourced',
    'database': madrigal_database,
    'facility': 'PFISR',
    'exp_name_pattern': [],
    'exp_check': False,
    'data_file_type': 'vi',
    'data_file_ext': ['h5', 'hdf5'],
    'data_root_dir': prf.datahub_data_root_dir / 'Madrigal' / 'PFISR',
    'allow_download': True,
    'status_control': False,
    'residual_control': False,
    'beam_location': True,
    'data_search_recursive': True,
    'label_fields': ['database', 'facility', 'site', 'data_file_type'],
}

default_variable_names = [
    'DATETIME', 'CGM_LAT', 'v_i_N', 'v_i_N_err', 'v_i_E', 'v_i_E_err', 'v_i_Z', 'v_i_Z_err', 'E_N', 'E_N_err', 'E_E',
    'E_E_err', 'GEO_ALT_MAX', 'GEO_ALT_MIN', 'INT_TIME',     'EF_MAG', 'EF_MAG_err', 'EF_ANGLE',  'EF_ANGLE_err',
    'v_i_MAG', 'v_i_MAG_err', 'v_i_ANGLE', 'v_i_ANGLE_err'
]

# default_data_search_recursive = True

default_attrs_required = []


class Dataset(datahub.DatasetSourced):
    def __init__(self, **kwargs):
        kwargs = basic.dict_set_default(kwargs, **default_dataset_attrs)

        super().__init__(**kwargs)

        self.database = kwargs.pop('database', '')
        self.facility = kwargs.pop('facility', '')
        self.site = kwargs.pop('site', PFISR('PFISR'))
        self.experiment = kwargs.pop('experiment', '')
        self.exp_ids = kwargs.pop('exp_ids', [])
        self.exp_name_pattern = kwargs.pop('exp_name_pattern', [])
        self.integration_time = kwargs.pop('integration_time', None)
        self.exp_check = kwargs.pop('exp_check', False)
        self.data_file_type = kwargs.pop('data_file_type', 'vi')
        self.affiliation = kwargs.pop('affiliation', '')
        self.allow_download = kwargs.pop('allow_download', True)
        self.metadata = None

        allow_load = kwargs.pop('allow_load', True)
        self.status_control = kwargs.pop('status_control', False)
        self.residual_control = kwargs.pop('residual_control', False)

        # self.config(**kwargs)

        if self.loader is None:
            self.loader = default_Loader
        if self.downloader is None:
            self.downloader = default_Downloader

        self._set_default_variables(
            default_variable_names,
            configured_variables=var_config.configured_variables
        )

        self._validate_attrs()

        if allow_load:
            self.load_data()

    def _validate_attrs(self):
        for attr_name in default_attrs_required:
            attr = getattr(self, attr_name)
            if not str(attr):
                mylog.StreamLogger.warning("The parameter {} is required before loading data!".format(attr_name))

        if self.exp_check:
            self.download_data(dry_run=True)

    def label(self, **kwargs):
        label = super().label()
        return label

    def load_data(self, **kwargs):
        self.check_data_files(**kwargs)

        for file_path in self.data_file_paths:
            load_obj = self.loader(file_path, )

            for var_name in self._variables.keys():
                self._variables[var_name].join(load_obj.variables[var_name])
            self.metadata = load_obj.metadata

        if self.time_clip:
            self.time_filter_by_range()

    def search_data_files(self, recursive=True, **kwargs):
        dt_fr = self.dt_fr
        dt_to = self.dt_to
        done = False
        if list(self.exp_ids):
            initial_file_dir = self.data_root_dir

            for exp_id in self.exp_ids:
                search_pattern = f"*EID-{exp_id}*/*{self.data_file_type}*"
                done = super().search_data_files(
                    initial_file_dir=initial_file_dir,
                    search_pattern=search_pattern, recursive=True, allow_multiple_files=True
                )

                if not done and self.allow_download:
                    done = self.download_data()
                    if done:
                        done = super().search_data_files(
                            initial_file_dir=initial_file_dir,
                            search_pattern=search_pattern, recursive=True, allow_multiple_files=True
                        )
                    else:
                        print('The requested experiment does not exist in the online database!')
                if len(done) > 1:
                    if isinstance(self.exp_name_pattern, (str, list)):
                        mylog.StreamLogger.warning(
                            "Multiple data files detected! Check the files:")
                    else:
                        mylog.StreamLogger.warning(
                            "Multiple data files detected!" +
                            "Specify the experiment name by the keyword 'exp_name_pattern' if possible.")
                    for fp in done:
                        mylog.simpleinfo.info(fp)

        else:
            diff_years = dt_to.year - dt_fr.year
            for ny in range(diff_years + 1):
                initial_file_dir = self.data_root_dir / str(dt_to.year + ny)
                search_pattern = "*EID-*/"
                exp_dirs = list(initial_file_dir.glob(search_pattern))

                if not list(exp_dirs) and self.allow_download:
                    self.download_data() 

                def dir_parser(dirs):
                    dirs_out = []
                    for d in dirs:
                        rc = re.compile(r'_([\d]{14})')
                        res = rc.findall(str(d))
                        dt_0 = datetime.datetime.strptime(res[0], "%Y%m%d%H%M%S")
                        dt_1 = datetime.datetime.strptime(res[1], "%Y%m%d%H%M%S")

                        if dt_1 < dt_fr:
                            continue
                        if dt_0 > dt_to:
                            continue
                        dirs_out.append(d)
                    return dirs_out

                file_dirs = dir_parser(exp_dirs)

                if not list(file_dirs) and self.allow_download:
                    self.download_data() 

                for fd in file_dirs:
                    if isinstance(self.exp_name_pattern, list):
                        p = '.*'.join(self.exp_name_pattern)
                    elif isinstance(self.exp_name_pattern, str):
                        p = self.exp_name_pattern
                    else:
                        p = '.*'
                    rc = re.compile(p)
                    res = rc.search(str(fd))
                    if res is None:
                        continue

                    file_patterns = ['PFISR', self.data_file_type]

                    # remove empty str
                    file_patterns = [pattern for pattern in file_patterns if str(pattern)]

                    search_pattern = '*' + '*'.join(file_patterns) + '*'

                    recursive = False

                    done = super().search_data_files(
                        initial_file_dir=fd, search_pattern=search_pattern,
                        allow_multiple_files=True, recursive=recursive)

                    # Validate file paths

                    if not done and self.allow_download:
                        done = self.download_data()
                        if done:
                            done = super().search_data_files(
                                initial_file_dir=initial_file_dir,
                                search_pattern=search_pattern,
                                allow_multiple_files=True, recursive=recursive
                            )
                        else:
                            print('The requested experiment does not exist in the online database!')

                    if len(done) > 1:
                        if isinstance(self.exp_name_pattern, (str, list)):
                            mylog.StreamLogger.warning(
                                "Multiple data files detected! Check the files:")
                        else:
                            mylog.StreamLogger.warning(
                                "Multiple data files detected!" +
                                "Specify the experiment name by the keyword 'exp_name_pattern' if possible.")
                        for fp in done:
                            mylog.simpleinfo.info(fp)

        self._select_file_by_time_resolution()
        import natsort
        self.data_file_paths = natsort.natsorted(self.data_file_paths, reverse=False)
        return done

    def _select_file_by_time_resolution(self):
        def show_option():
            mylog.simpleinfo.info("Options for the integration time [s]: ")
            ts = np.unique(itg_times)
            for t in ts:
                mylog.simpleinfo.info(f"{t}")
                
        file_paths = self.data_file_paths
        itg_times = []
        for file_path in file_paths:
            with h5py.File(file_path, 'r') as fh5:
                itg_time = np.array(fh5['Data']['Array Layout']['1D Parameters']['inttms']).astype(np.float32)
                itg_times.append(np.median(itg_time))

        if self.integration_time is None:
            mylog.StreamLogger.error(f"The integration time was not set. Reset is required!")
            show_option()
            raise AttributeError
            
        if self.integration_time not in itg_times:
            mylog.StreamLogger.error(f"The integration time is not found. Reset is required!")
            show_option()
            raise AttributeError 
        
        ind = np.where(np.array(itg_times, dtype=np.float32) == self.integration_time)[0]

        file_paths = [file_paths[i] for i in ind]
        mylog.simpleinfo.info(f'The following files are selected with the integration time {self.integration_time} [s]:')
        for file_path in file_paths:
            mylog.simpleinfo.info(file_path)

        self.data_file_paths = [self.data_file_paths[i] for i in ind]

    def download_data(self, dry_run=False):
        include_exp_name_patterns = [self.exp_name_pattern] if isinstance(self.exp_name_pattern, list) else []
        download_obj = self.downloader(
            dt_fr=self.dt_fr, dt_to=self.dt_to,
            dry_run=dry_run,
            data_file_root_dir=self.data_root_dir,
            include_exp_ids=self.exp_ids,
            include_exp_name_patterns=include_exp_name_patterns)
        return download_obj.data_file_paths


    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        if isinstance(value, str):
            self._database = DatabaseModel(value)
        elif issubclass(value.__class__, DatabaseModel):
            self._database = value
        else:
            raise TypeError

    @property
    def facility(self):
        return self._facility

    @facility.setter
    def facility(self, value):
        if isinstance(value, str):
            self._facility = FacilityModel(value)
        elif issubclass(value.__class__, FacilityModel):
            self._facility = value
        else:
            raise TypeError

    @property
    def site(self):
        return self._site

    @site.setter
    def site(self, value):
        if isinstance(value, str):
            self._site = PFISR(value)
        elif issubclass(value.__class__, SiteModel):
            self._site = value
        else:
            raise TypeError


class PFISR(SiteModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj

    def __init__(self, str_in, **kwargs):
        self.name = 'Poker Flat ISR'
        self.location = {
            'GEO_LAT': 65.13,
            'GEO_LON': 212.529,
            'GEO_ALT': 0.215
        }















# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import pathlib

import requests
import bs4
import re
import tqdm

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.config import prf


class Downloader(object):

    def __init__(
        self, 
        dt_fr, 
        dt_to,
        direct_download=True, 
        force_download=False,
        data_file_root_dir=None,
        dry_run=False,
        from_ftp=False,
        ):

        self.url_base = "https://cdaweb.gsfc.nasa.gov/pub/data/"
        self.force_download = force_download
        self.dry_run=dry_run
        self.from_ftp=from_ftp

        self.dt_fr = dt_fr
        self.dt_to = dt_to
        self.source_file_paths = []
        self.done = False

        if data_file_root_dir is None:
            self.data_file_root_dir = prf.datahub_data_root_dir
        else:
            self.data_file_root_dir = data_file_root_dir

        if direct_download:
            self.download()

    def download(self,):
        if self.from_ftp:
            raise NotImplementedError
        else:
            self.download_from_http()
            
    def search_from_http(self, subdirs=None, file_name_patterns=None):
        url = self.url_base + '/'.join(subdirs)
        r = requests.get(url)

        soup = bs4.BeautifulSoup(r.text, 'html.parser')
        a_tags = soup.find_all('a', href=True)
        hrefs = [a_tag['href'] for a_tag in a_tags]

        search_pattern = '.*' + '.*'.join(file_name_patterns) + '.*'
        fn_regex = re.compile(search_pattern)
        hrefs = list(filter(fn_regex.match, hrefs))

        paths = []
        for href in hrefs:
            paths.append(url + '/' + href)
        return paths

    def download_from_http(self, ):

        source_file_paths = self.search_from_http()

        for url  in source_file_paths:
            if self.dry_run:
                print(f"Dry run: {url}.")
            else:
                self.save_file_from_http(url=url)

    def save_file_from_http(self, url, file_dir=None, file_name=None):
            if file_name is None:
                file_name = url.split('/')[-1]
            file_path = file_dir / file_name
            if file_path.is_file():
                mylog.simpleinfo.info(
                    "The file {} exists in the directory {}.".format(file_path.name, file_path.parent.resolve()))
                if not self.force_download:
                    self.done = True
                    return

            file_path.parent.resolve().mkdir(parents=True, exist_ok=True)
            mylog.simpleinfo.info(f'Downloading {file_name} ...')
            res = self._download_by_requests_get(url, data_file_path=file_path)
            if res:
                mylog.simpleinfo.info(f'Saved in {file_dir}')
                self.done = True
            else:
                mylog.StreamLogger.error(f"Error during downloading. Code: {res}.")

            return

    @staticmethod
    def _download_by_requests_get(url, data_file_path=None, params=None, stream=True, allow_redirects=True, **kwargs):
        r = requests.get(url, params=params, stream=stream, allow_redirects=allow_redirects, **kwargs)
        if r.status_code != 200:
            return -2

        total_size_in_bytes = int(r.headers.get('content-length', 0))
        block_size = 1024  # 1 Kibibyte
        progress_bar = tqdm.tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)

        with open(data_file_path, 'wb') as file:
            for data in r.iter_content(block_size):
                progress_bar.update(len(data))
                file.write(data)
        progress_bar.close()
        if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
            mylog.StreamLogger.error("Something wrong during the download!")
            data_file_path.unlink(missing_ok=True)
            return -1

        return 1

    def search_from_ftp(self, ftp, file_name_patterns=None):
        paths = []

        return paths

    def save_file(self, file_dir, file_name, r_source_file):
        file_path = file_dir / file_name

        if file_path.is_file():
            mylog.simpleinfo.info(
                "The file {} exists in the directory {}.".format(file_path.name, file_path.parent.resolve()))

        if self.force_download:
            file_path.parent.resolve().mkdir(parents=True, exist_ok=True)
            with open(file_path, "wb") as f:
                mylog.simpleinfo.info(
                    "Downloading {} to the directory {} ...".format(file_path.name, file_path.parent.resolve())
                )
                f.write(r_source_file.content)
                mylog.simpleinfo.info("Done")




# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

import netCDF4
import datetime
import numpy as np
import geospacelab.toolbox.utilities.pydatetime as dttool


class Loader(object):

    def __init__(self, file_path, file_type='fitted', pole='N'):

        self.variables = {}
        self.metadata = {}
        self.file_path = file_path
        self.file_type = file_type
        self.pole = pole

        self.load_data()

    def load_data(self):
        if self.file_type == 'fitted':
            self.load_data_fitted()

    def load_data_fitted(self):
        dataset = netCDF4.Dataset(self.file_path)
        variables = {}

        time_1 = np.array([
            datetime.datetime(yy, mm, dd, HH, MM, SS)
            for yy, mm, dd, HH, MM, SS in zip(
                dataset.variables['start_yr'][::],
                dataset.variables['start_mo'][::],
                dataset.variables['start_dy'][::],
                dataset.variables['start_hr'][::],
                dataset.variables['start_mt'][::],
                dataset.variables['start_sc'][::],
            )
        ])
        time_2 = np.array([
            datetime.datetime(yy, mm, dd, HH, MM, SS)
            for yy, mm, dd, HH, MM, SS in zip(
                dataset.variables['end_yr'][::],
                dataset.variables['end_mo'][::],
                dataset.variables['end_dy'][::],
                dataset.variables['end_hr'][::],
                dataset.variables['end_mt'][::],
                dataset.variables['end_sc'][::],
            )
        ])

        ntime = time_1.shape[0]
        nlon = dataset.variables['nlon'][0]
        nlat = dataset.variables['nlat'][0]
        colat = dataset.variables['colat'][::].reshape((ntime, nlon, nlat))

        mlt = dataset.variables['mlt'][::].reshape((ntime, nlon, nlat))

        Jr = dataset.variables['Jr'][::].reshape(ntime, nlon, nlat)

        variables['DATETIME'] = np.reshape(time_1 + (time_2 - time_1) / 2, (ntime, 1))
        variables['DATETIME_1'] = np.reshape(time_1, (ntime, 1))
        variables['DATETIME_2'] = np.reshape(time_2, (ntime, 1))

        variables['GRID_MLAT'] = np.array(90. - colat)
        variables['GRID_MLT'] = np.array(mlt)

        variables['GRID_Jr'] = np.array(Jr)

        dataset.close()

        self.variables = variables


if __name__ == "__main__":
    import pathlib
    fp = pathlib.Path('/home/lei/afys-data/JHUAPL/AMPERE/Fitted/201603/AMPERE_fitted_20160314.0000.86400.600.north.grd.ncdf')
    loader = Loader(file_path=fp)


    # if hasattr(readObj, 'pole'):
    #    readObj.filter_data_pole(boundinglat = 25)